#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2014, Simon Harrison
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from shapely.geometry import MultiPolygon, Point


def parse_lat_long(coords):
    # WKT standard is: POINT (x y)
    # WSG 84: http://spatialreference.org/ref/epsg/4326/
    lat, lon = coords
    point = Point(float(lat), float(lon))
    return point


def parse_poly(lines):
    """ Parse an Osmosis/Google polygon filter file.

    :Params:
        lines : open file pointer

    :Returns:
        shapely.geometry.MultiPolygon object.

    .. note::
        http://wiki.openstreetmap.org/wiki/Osmosis/Polygon_Filter_File_Format

    """
    in_ring = False
    coords = []

    for index, line in enumerate(lines):
        if index == 0:
            # ignore meta/garbage.
            continue
        elif index == 1:
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True
        elif in_ring and line.strip() == 'END':
            in_ring = False
        elif in_ring:
            ring.append(map(float, line.split()))
        elif not in_ring and line.strip() == 'END':
            break
        elif not in_ring and line.startswith('!'):
            coords[-1][1].append([])
            ring = coords[-1][1][-1]
            in_ring = True
        elif not in_ring:
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True

    return MultiPolygon(coords)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping


class KeyValueList(list):
    """ A KeyValueList is a list of key-value pairs that functions as an
    ordered dictionary with support for duplicate keys.

    An instance can be created using either an iterable sequence of pairs or a
    mapping plus an optional set of keyword arguments:

    >>> kvl = KeyValueList([('a', 1), ('c', 7), 'cx', ('b', [8, 9])], c='b')
    >>> kvl
    KeyValueList([('a', 1), ('c', 7), ('c', 'x'), ('b', [8, 9]), ('c', 'b')])

    This creates a list of values that have both an associated key and a
    positional index:

     index | key | value
    -------+-----+--------
     0     | 'a' | 1
     1     | 'c' | 7
     2     | 'c' | 1
     3     | 'b' | [8, 9]
     4     | 'c' | 'b'

    """

    def __init__(self, iterable=(), **kwargs):
        list.__init__(self)
        self.extend(iterable)
        self.extend(kwargs)

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, list.__repr__(self))

    def __getitem__(self, index):
        """ Get a single item.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl[2]
        ('green', 'grass')
        >>> kvl[9]
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        IndexError: list index out of range

        """
        try:
            got = list.__getitem__(self, index)
        except TypeError:
            try:
                return next(self.get(index))
            except StopIteration:
                return None
        else:
            if isinstance(index, slice):
                return KeyValueList(got)
            else:
                return got

    def __getslice__(self, start, end):
        """ Get a slice of items.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl[1:3]
        KeyValueList([('blue', 'sea'), ('green', 'grass')])

        """
        try:
            return KeyValueList(list.__getslice__(self, start, end))
        except AttributeError:
            return KeyValueList(list.__getitem__(self, slice(start, end)))

    def __setitem__(self, index, item):
        """ Set a single item.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl[1] = ('yellow', 'sun')
        >>> for item in kvl:
        ...     print(item)
        ('red', 'rose')
        ('yellow', 'sun')
        ('green', 'grass')
        ('blue', 'sky')
        >>> kvl[9] = ('orange', 'orange')
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        IndexError: list assignment index out of range

        """
        list.__setitem__(self, index, item)

    def __setslice__(self, start, stop, items):
        """ Set a slice of items.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl[1:3] = [('yellow', 'sun'), ('grey', 'stone'), ('red', 'berry')]
        >>> for item in kvl:
        ...     print(item)
        ('red', 'rose')
        ('yellow', 'sun')
        ('grey', 'stone')
        ('red', 'berry')
        ('blue', 'sky')

        """
        list.__setslice__(self, start, stop, items)

    def __delitem__(self, index):
        """ Delete a single item.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> del kvl[1]
        >>> for item in kvl:
        ...     print(item)
        ('red', 'rose')
        ('green', 'grass')
        ('blue', 'sky')
        >>> del kvl[9]
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        IndexError: list assignment index out of range

        """
        list.__delitem__(self, index)

    def __delslice__(self, start, end):
        """ Delete a slice of items.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> del kvl[1:3]
        >>> for item in kvl:
        ...     print(item)
        ('red', 'rose')
        ('blue', 'sky')

        """
        list.__delslice__(self, start, end)

    def __contains__(self, item):
        """ Check for the presence of a particular key-value pair within this
        list. This is equivalent to the `has_item` method.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> ('blue', 'sea') in kvl
        True
        >>> ('yellow', 'sun') in kvl
        False
        >>> ('purple', 'grape') not in kvl
        True
        >>> ('red', 'rose') not in kvl
        False

        """
        key, value = item
        return self.has_item(key, value)

    def __iter__(self):
        """ Iterate through all items in this list. This is equivalent to the
        `items` method with `collect` set to False.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> for key, value in kvl:
        ...     print("{0} -> {1}".format(key, value))
        red -> rose
        blue -> sea
        green -> grass
        blue -> sky

        """
        return list.__iter__(self)

    def append(self, key, value):
        """ Append a single key-value pair to the end of this list.

        >>> kvl = KeyValueList()
        >>> kvl.append('one', 'eins')
        >>> kvl.append('two', 'zwei')
        >>> kvl.append('three', 'drei')
        >>> for item in kvl:
        ...     print(item)
        ('one', 'eins')
        ('two', 'zwei')
        ('three', 'drei')

        """
        list.append(self, (key, value))

    def extend(self, iterable):
        """ Concatenate two lists by adding a list of extra items to the end
        of this list. Each item added must be capable of being unpacked into a
        key-value pair.

        >>> kvl = KeyValueList([('one', 'eins'), ('two', 'zwei')])
        >>> kvl.extend([('three', 'drei'), ('four', 'vier')])
        >>> for item in kvl:
        ...     print(item)
        ('one', 'eins')
        ('two', 'zwei')
        ('three', 'drei')
        ('four', 'vier')
        >>> kvl.extend(['five', 'six'])
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        ValueError: KeyValueList items must be pairs

        """
        if isinstance(iterable, Mapping):
            list.extend(self, iterable.items())
        else:
            try:
                list.extend(self, ((k, v) for k, v in iterable))
            except ValueError:
                raise ValueError("KeyValueList items must be pairs")

    def insert(self, index, key, value):
        """ Insert a key-value pair at a particular position within the list:

        >>> kvl = KeyValueList({'three': 'drei'})
        >>> kvl.insert(0, 'one', 'eins')
        >>> kvl.insert(1, 'two', 'zwei')
        >>> for item in kvl:
        ...     print(item)
        ('one', 'eins')
        ('two', 'zwei')
        ('three', 'drei')

        """
        list.insert(self, index, (key, value))

    def has_item(self, key, value):
        """ Check for the presence of a particular key-value pair within this
        list.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl.has_item('green', 'grass')
        True
        >>> kvl.has_item('pink', 'rose')
        False

        """
        for k, v in self:
            if k == key and v == value:
                return True
        return False

    def has_key(self, key):
        """ Check for the presence of a particular key within this list.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl.has_key('blue')
        True
        >>> kvl.has_key('yellow')
        False

        """
        for k, v in self:
            if k == key:
                return True
        return False

    def has_value(self, value):
        """ Check for the presence of a particular value within this list.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl.has_value('grass')
        True
        >>> kvl.has_value('sun')
        False

        """
        for k, v in self:
            if v == value:
                return True
        return False

    def get(self, key):
        """ Iterate through all values associated with the specified key.
        Non-existent keys return empty iterators.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> for value in kvl.get('blue'):
        ...     print(value)
        sea
        sky
        >>> for value in kvl.get('yellow'):
        ...     print(value)

        """
        return (v for k, v in self if k == key)

    def put(self, key, *values):
        """ Replace all values associated with the specified key with the
        value arguments provided. If fewer values are specified than currently
        exist, remaining items will be removed; if more items are specified,
        these will be appended to the end of the list.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('blue', 'sky')])
        >>> kvl.put('blue', 'jeans')
        >>> kvl
        KeyValueList([('red', 'rose'), ('blue', 'jeans')])
        >>> kvl.put('red', 'heart', 'berry')
        >>> kvl
        KeyValueList([('red', 'heart'), ('blue', 'jeans'), ('red', 'berry')])

        """
        new_values = list(values)
        self[:] = [
            (k, value) if k != key else (k, new_values.pop(0))
            for k, value in self
            if k != key or new_values
        ]
        self.extend([(key, value) for value in new_values])

    def remove(self, key):
        """ Remove all items from this list that contain the key specified. If
        the key is not found, a ValueError is raised.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl.remove('blue')
        >>> kvl
        KeyValueList([('red', 'rose'), ('green', 'grass')])
        >>> kvl.remove('yellow')
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        ValueError: Key 'yellow' not in list

        """
        length = len(self)
        self[:] = ((k, v) for k, v in self if k != key)
        if len(self) == length:
            raise ValueError("Key {0} not in list".format(repr(key)))

    def pop(self, index=None):
        """ Remove the item at the index specified and return it. If no index
        is specified, the last item is popped. If the index is out of range, an
        IndexError is raised.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> kvl.pop(2)
        ('green', 'grass')
        >>> kvl.pop()
        ('blue', 'sky')
        >>> kvl.pop(6)
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        IndexError: pop index out of range

        """
        if index is None:
            return list.pop(self)
        else:
            return list.pop(self, index)

    def clear(self):
        """ Remove all items from this list.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> len(kvl)
        4
        >>> kvl.clear()
        >>> len(kvl)
        0

        """
        del self[:]

    def sort(self, *args, **kwargs):
        """ Sort the items in this list into ascending order.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> for item in kvl:
        ...     print(item)
        ('red', 'rose')
        ('blue', 'sea')
        ('green', 'grass')
        ('blue', 'sky')
        >>> kvl.sort()
        >>> for item in kvl:
        ...     print(item)
        ('blue', 'sea')
        ('blue', 'sky')
        ('green', 'grass')
        ('red', 'rose')

        """
        list.sort(self, *args, **kwargs)

    def reverse(self):
        """ Reverse the order of items in this list.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> for item in kvl:
        ...     print(item)
        ('red', 'rose')
        ('blue', 'sea')
        ('green', 'grass')
        ('blue', 'sky')
        >>> kvl.reverse()
        >>> for item in kvl:
        ...     print(item)
        ('blue', 'sky')
        ('green', 'grass')
        ('blue', 'sea')
        ('red', 'rose')

        """
        list.reverse(self)

    def copy(self):
        """ Create and return a shallow copy of this list instance.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea')])
        >>> kvl
        KeyValueList([('red', 'rose'), ('blue', 'sea')])
        >>> kvl2 = kvl.copy()
        >>> kvl2.append('green', 'grass')
        >>> kvl
        KeyValueList([('red', 'rose'), ('blue', 'sea')])
        >>> kvl2
        KeyValueList([('red', 'rose'), ('blue', 'sea'), ('green', 'grass')])

        """
        return KeyValueList(self)

    def iterkeys(self, collect=False):
        """ Iterate through the keys in this list. If `collect` is True,
        yield only unique keys.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> for key in kvl.iterkeys():
        ...     print(key)
        red
        blue
        green
        blue
        >>> for key in kvl.iterkeys(collect=True):
        ...     print(key)
        red
        blue
        green

        """
        if collect:
            keys = []
            for k, v in self:
                if k not in keys:
                    keys.append(k)
                    yield k
        else:
            for k, v in self:
                yield k

    def itervalues(self, collect=False):
        """ Iterate through the values in this list. If `collect` is True,
        yield a list of values for each unique key.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> for key in kvl.itervalues():
        ...     print(key)
        rose
        sea
        grass
        sky
        >>> for key in kvl.itervalues(collect=True):
        ...     print(key)
        ['rose']
        ['sea', 'sky']
        ['grass']

        """
        if collect:
            keys, values = [], []
            for k, v in self:
                try:
                    index = keys.index(k)
                except ValueError:
                    keys.append(k)
                    values.append([v])
                else:
                    values[index].append((v))
            for value in values:
                yield value
        else:
            for k, v in self:
                yield v

    def iteritems(self, collect=False):
        """ Iterate through the items in this list. If `collect` is True,
        yield only one item for each unique key, each with a list of
        associated values.

        >>> kvl = KeyValueList([('red', 'rose'), ('blue', 'sea'),
        ... ('green', 'grass'), ('blue', 'sky')])
        >>> for key in kvl.iteritems():
        ...     print(key)
        ('red', 'rose')
        ('blue', 'sea')
        ('green', 'grass')
        ('blue', 'sky')
        >>> for key in kvl.iteritems(collect=True):
        ...     print(key)
        ('red', ['rose'])
        ('blue', ['sea', 'sky'])
        ('green', ['grass'])

        """
        if collect:
            keys, items = [], []
            for k, v in self:
                try:
                    index = keys.index(k)
                except ValueError:
                    keys.append(k)
                    items.append((k, [v]))
                else:
                    items[index][1].append((v))
            for item in items:
                yield item
        else:
            for item in self:
                yield item



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

from py2neo.batch import WriteBatch
from py2neo.core import Node
from py2neo.cypher.util import StartOrMatch


__all__ = ["Store", "NotSaved"]


class NotSaved(ValueError):
    """ Raised when an object has not been saved but a bound node is required.
    """
    pass


class Store(object):
    """ Virtual storage mapped onto an existing graph in which
    objects can be stored.
    """

    def __init__(self, graph):
        self.graph = graph
        if self.graph.supports_optional_match:
            self.__delete_query = (
                StartOrMatch(self.graph).node("a", "{A}").string +
                "OPTIONAL MATCH a-[r]-b "
                "DELETE r, a"
            )
        else:
            self.__delete_query = (
                StartOrMatch(self.graph).node("a", "{A}").string +
                "MATCH a-[r?]-b "
                "DELETE r, a"
            )

    def _assert_saved(self, subj):
        try:
            node = subj.__node__
            if node is None:
                raise NotSaved(subj)
        except AttributeError:
            raise NotSaved(subj)

    def _get_node(self, endpoint):
        if isinstance(endpoint, Node):
            return endpoint
        if not hasattr(endpoint, "__node__"):
            self.save(endpoint)
        return endpoint.__node__

    def _is_same(self, obj, endpoint):
        if isinstance(endpoint, Node):
            if hasattr(obj, "__node__"):
                return endpoint == obj.__node__
            else:
                return False
        else:
            return endpoint is obj

    def is_saved(self, subj):
        """ Return :py:const:`True` if the object `subj` has been saved to
        the database, :py:const:`False` otherwise.

        :param subj: the object to test
        """
        return hasattr(subj, "__node__") and subj.__node__ is not None

    def relate(self, subj, rel_type, obj, properties=None):
        """ Define a relationship between `subj` and `obj` of type `rel_type`.
        This is a local operation only: nothing is saved to the database until
        a save method is called. Relationship properties may optionally be
        specified.

        :param subj: the object bound to the start of the relationship
        :param rel_type: the relationship type
        :param obj: the object bound to the end of the relationship
        :param properties: properties attached to the relationship (optional)
        """
        if not hasattr(subj, "__rel__"):
            subj.__rel__ = {}
        if rel_type not in subj.__rel__:
            subj.__rel__[rel_type] = []
        subj.__rel__[rel_type].append((properties or {}, obj))

    def separate(self, subj, rel_type, obj=None):
        """ Remove any relationship definitions which match the criteria
        specified. This is a local operation only: nothing is saved to the
        database until a save method is called. If no object is specified, all
        relationships of type `rel_type` are removed.

        :param subj: the object bound to the start of the relationship
        :param rel_type: the relationship type
        :param obj: the object bound to the end of the relationship (optional)
        """
        if not hasattr(subj, "__rel__"):
            return
        if rel_type not in subj.__rel__:
            return
        if obj is None:
            del subj.__rel__[rel_type]
        else:
            subj.__rel__[rel_type] = [
                (props, endpoint)
                for props, endpoint in subj.__rel__[rel_type]
                if not self._is_same(obj, endpoint)
            ]

    def load_related(self, subj, rel_type, cls):
        """ Load all nodes related to `subj` by a relationship of type
        `rel_type` into objects of type `cls`.

        :param subj: the object bound to the start of the relationship
        :param rel_type: the relationship type
        :param cls: the class to load all related objects into
        :return: list of `cls` instances
        """
        if not hasattr(subj, "__rel__"):
            return []
        if rel_type not in subj.__rel__:
            return []
        return [
            self.load(cls, self._get_node(endpoint))
            for rel_props, endpoint in subj.__rel__[rel_type]
        ]

    def load(self, cls, node):
        """ Load and return an object of type `cls` from database node `node`.

        :param cls: the class of the object to be returned
        :param node: the node from which to load object data
        :return: a `cls` instance
        """
        subj = cls()
        setattr(subj, "__node__", node)
        self.reload(subj)
        return subj

    def load_indexed(self, index_name, key, value, cls):
        """ Load zero or more indexed nodes from the database into a list of
        objects.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param cls: the class of the object to be returned
        :return: a list of `cls` instances
        """
        index = self.graph.legacy.get_index(Node, index_name)
        nodes = index.get(key, value)
        return [self.load(cls, node) for node in nodes]

    def load_unique(self, index_name, key, value, cls):
        """ Load a uniquely indexed node from the database into an object.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param cls: the class of the object to be returned
        :return: as instance of `cls` containing the loaded data
        """
        index = self.graph.legacy.get_index(Node, index_name)
        nodes = index.get(key, value)
        if not nodes:
            return None
        if len(nodes) > 1:
            raise LookupError("Multiple nodes match the given criteria; "
                              "consider using `load_all` instead.")
        return self.load(cls, nodes[0])

    def reload(self, subj):
        """ Reload properties and relationships from a database node into
        `subj`.

        :param subj: the object to reload
        :raise NotSaved: if `subj` is not linked to a database node
        """
        self._assert_saved(subj)
        # naively copy properties from node to object
        properties = subj.__node__.get_properties()
        for key in subj.__dict__:
            if not key.startswith("_") and key not in properties:
                setattr(subj, key, None)
        for key, value in properties.items():
            if not key.startswith("_"):
                setattr(subj, key, value)
        subj.__rel__ = {}
        for rel in subj.__node__.match():
            if rel.type not in subj.__rel__:
                subj.__rel__[rel.type] = []
            subj.__rel__[rel.type].append((rel.get_properties(), rel.end_node))

    def save(self, subj, node=None):
        """ Save an object to a database node.

        :param subj: the object to save
        :param node: the database node to save to (if omitted, will re-save to
            same node as previous save)
        """
        if node is not None:
            subj.__node__ = node
        # naively copy properties from object to node
        props = {}
        for key, value in subj.__dict__.items():
            if not key.startswith("_"):
                props[key] = value
        if hasattr(subj, "__node__"):
            subj.__node__.set_properties(props)
            self.graph.cypher.run(StartOrMatch(self.graph).node("a", "{a}").string +
                                  "MATCH (a)-[r]->(b) DELETE r",
                                  {"a": subj.__node__})
        else:
            subj.__node__, = self.graph.create(props)
        # write rels
        if hasattr(subj, "__rel__"):
            batch = WriteBatch(self.graph)
            for rel_type, rels in subj.__rel__.items():
                for rel_props, endpoint in rels:
                    end_node = self._get_node(endpoint)
                    if end_node not in self.graph:
                        raise ValueError(end_node)
                    batch.create((subj.__node__, rel_type, end_node, rel_props))
            batch.run()
        return subj

    def save_indexed(self, index_name, key, value, *subj):
        """ Save one or more objects to the database, indexed under the
        supplied criteria.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param subj: one or more objects to save
        """
        index = self.graph.legacy.get_or_create_index(Node, index_name)
        for subj in subj:
            index.add(key, value, self.save(self._get_node(subj)))

    def save_unique(self, index_name, key, value, subj):
        """ Save an object to the database, uniquely indexed under the
        supplied criteria.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param subj: the object to save
        """
        index = self.graph.legacy.get_or_create_index(Node, index_name)
        node = index.get_or_create(key, value, {})
        self.save(subj, node)

    def delete(self, subj):
        """ Delete a saved object node from the database as well as all
        incoming and outgoing relationships.

        :param subj: the object to delete from the database
        :raise NotSaved: if `subj` is not linked to a database node
        """
        self._assert_saved(subj)
        node = subj.__node__
        del subj.__node__
        self.graph.cypher.execute(self.__delete_query, {"A": node})



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.legacy.batch import LegacyWriteBatch
from py2neo.core import Service, Node, Relationship, Resource, ResourceTemplate
from py2neo.packages.jsonstream import assembled, grouped
from py2neo.packages.httpstream.numbers import CREATED
from py2neo.packages.httpstream.packages.urimagic import percent_encode, URI


__all__ = ["Index"]


class Index(Service):
    """ Searchable database index which can contain either nodes or
    relationships.

    .. seealso:: :py:func:`Graph.get_or_create_index`
    """

    __instances = {}

    def __new__(cls, content_type, uri, name=None):
        """ Fetch a cached instance if one is available, otherwise create,
        cache and return a new instance.

        :param uri: URI of the cached resource
        :return: a resource instance
        """
        inst = super(Index, cls).__new__(cls)
        return cls.__instances.setdefault(uri, inst)

    def __init__(self, content_type, uri, name=None):
        Service.__init__(self)
        self._content_type = content_type
        key_value_pos = uri.find("/{key}/{value}")
        if key_value_pos >= 0:
            self._searcher = ResourceTemplate(uri)
            self.bind(uri[:key_value_pos])
        else:
            self.bind(uri)
            self._searcher = ResourceTemplate(uri.string + "/{key}/{value}")
        uri = self.resource.uri
        if self.graph.neo4j_version >= (1, 9):
            self._create_or_fail = Resource(uri.resolve("?uniqueness=create_or_fail"))
            self._get_or_create = Resource(uri.resolve("?uniqueness=get_or_create"))
        else:
            self._create_or_fail = None
            self._get_or_create = Resource(uri.resolve("?unique"))
        self._query_template = ResourceTemplate(uri.string + "{?query,order}")
        self._name = name or uri.path.segments[-1]
        self.__searcher_stem_cache = {}

    def __repr__(self):
        return "{0}({1}, {2})".format(
            self.__class__.__name__,
            self._content_type.__name__,
            repr(self.uri.string)
        )

    def _searcher_stem_for_key(self, key):
        if key not in self.__searcher_stem_cache:
            stem = self._searcher.uri_template.string.partition("{key}")[0]
            self.__searcher_stem_cache[key] = stem + percent_encode(key) + "/"
        return self.__searcher_stem_cache[key]

    def add(self, key, value, entity):
        """ Add an entity to this index under the `key`:`value` pair supplied::

            # create a node and obtain a reference to the "People" node index
            alice, = graph.create({"name": "Alice Smith"})
            people = graph.get_or_create_index(neo4j.Node, "People")

            # add the node to the index
            people.add("family_name", "Smith", alice)

        Note that while Neo4j indexes allow multiple entities to be added under
        a particular key:value, the same entity may only be represented once;
        this method is therefore idempotent.
        """
        self.resource.post({
            "key": key,
            "value": value,
            "uri": entity.uri.string,
        })
        return entity

    def add_if_none(self, key, value, entity):
        """ Add an entity to this index under the `key`:`value` pair
        supplied if no entry already exists at that point::

            # obtain a reference to the "Rooms" node index and
            # add node `alice` to room 100 if empty
            rooms = graph.get_or_create_index(neo4j.Node, "Rooms")
            rooms.add_if_none("room", 100, alice)

        If added, this method returns the entity, otherwise :py:const:`None`
        is returned.
        """
        rs = self._get_or_create.post({
            "key": key,
            "value": value,
            "uri": entity.uri.string,
        })
        if rs.status_code == CREATED:
            return entity
        else:
            return None

    @property
    def content_type(self):
        """ Return the type of entity contained within this index. Will return
        either :py:class:`Node` or :py:class:`Relationship`.
        """
        return self._content_type

    @property
    def name(self):
        """ Return the name of this index.
        """
        return self._name

    def get(self, key, value):
        """ Fetch a list of all entities from the index which are associated
        with the `key`:`value` pair supplied::

            # obtain a reference to the "People" node index and
            # get all nodes where `family_name` equals "Smith"
            people = graph.get_or_create_index(neo4j.Node, "People")
            smiths = people.get("family_name", "Smith")

        ..
        """
        return [
            self.graph.hydrate(assembled(result))
            for i, result in grouped(self._searcher.expand(key=key, value=value).get())
        ]

    def create(self, key, value, abstract):
        """ Create and index a new node or relationship using the abstract
        provided.
        """
        batch = LegacyWriteBatch(self.graph)
        if self._content_type is Node:
            batch.create(abstract)
            batch.add_to_index(Node, self, key, value, 0)
        elif self._content_type is Relationship:
            batch.create(abstract)
            batch.add_to_index(Relationship, self, key, value, 0)
        else:
            raise TypeError(self._content_type)
        entity, index_entry = batch.submit()
        return entity

    def _create_unique(self, key, value, abstract):
        """ Internal method to support `get_or_create` and `create_if_none`.
        """
        if self._content_type is Node:
            body = {
                "key": key,
                "value": value,
                "properties": abstract
            }
        elif self._content_type is Relationship:
            body = {
                "key": key,
                "value": value,
                "start": abstract[0].uri.string,
                "type": abstract[1],
                "end": abstract[2].uri.string,
                "properties": abstract[3] if len(abstract) > 3 else None
            }
        else:
            raise TypeError(self._content_type)
        return self._get_or_create.post(body)

    def get_or_create(self, key, value, abstract):
        """ Fetch a single entity from the index which is associated with the
        `key`:`value` pair supplied, creating a new entity with the supplied
        details if none exists::

            # obtain a reference to the "Contacts" node index and
            # ensure that Alice exists therein
            contacts = graph.get_or_create_index(neo4j.Node, "Contacts")
            alice = contacts.get_or_create("name", "SMITH, Alice", {
                "given_name": "Alice Jane", "family_name": "Smith",
                "phone": "01234 567 890", "mobile": "07890 123 456"
            })

            # obtain a reference to the "Friendships" relationship index and
            # ensure that Alice and Bob's friendship is registered (`alice`
            # and `bob` refer to existing nodes)
            friendships = graph.get_or_create_index(neo4j.Relationship, "Friendships")
            alice_and_bob = friendships.get_or_create(
                "friends", "Alice & Bob", (alice, "KNOWS", bob)
            )

        ..
        """
        return self.graph.hydrate(assembled(self._create_unique(key, value, abstract)))

    def create_if_none(self, key, value, abstract):
        """ Create a new entity with the specified details within the current
        index, under the `key`:`value` pair supplied, if no such entity already
        exists. If creation occurs, the new entity will be returned, otherwise
        :py:const:`None` will be returned::

            # obtain a reference to the "Contacts" node index and
            # create a node for Alice if one does not already exist
            contacts = graph.get_or_create_index(neo4j.Node, "Contacts")
            alice = contacts.create_if_none("name", "SMITH, Alice", {
                "given_name": "Alice Jane", "family_name": "Smith",
                "phone": "01234 567 890", "mobile": "07890 123 456"
            })

        ..
        """
        rs = self._create_unique(key, value, abstract)
        if rs.status_code == CREATED:
            return self.graph.hydrate(assembled(rs))
        else:
            return None

    def remove(self, key=None, value=None, entity=None):
        """ Remove any entries from the index which match the parameters
        supplied. The allowed parameter combinations are:

        `key`, `value`, `entity`
            remove a specific entity indexed under a given key-value pair

        `key`, `value`
            remove all entities indexed under a given key-value pair

        `key`, `entity`
            remove a specific entity indexed against a given key but with
            any value

        `entity`
            remove all occurrences of a specific entity regardless of
            key and value

        """
        if key and value and entity:
            t = ResourceTemplate(self.resource.uri.string + "/{key}/{value}/{entity}")
            t.expand(key=key, value=value, entity=entity._id).delete()
        elif key and value:
            uris = [
                URI(entity.resource.metadata["indexed"])
                for entity in self.get(key, value)
            ]
            batch = LegacyWriteBatch(self.graph)
            for uri in uris:
                batch.append_delete(uri)
            batch.run()
        elif key and entity:
            t = ResourceTemplate(self.resource.uri.string + "/{key}/{entity}")
            t.expand(key=key, entity=entity._id).delete()
        elif entity:
            t = ResourceTemplate(self.resource.uri.string + "/{entity}")
            t.expand(entity=entity._id).delete()
        else:
            raise TypeError("Illegal parameter combination for index removal")

    def query(self, query):
        """ Query the index according to the supplied query criteria, returning
        a list of matched entities::

            # obtain a reference to the "People" node index and
            # get all nodes where `family_name` equals "Smith"
            people = graph.get_or_create_index(neo4j.Node, "People")
            s_people = people.query("family_name:S*")

        The query syntax used should be appropriate for the configuration of
        the index being queried. For indexes with default configuration, this
        should be Apache Lucene query syntax.
        """
        resource = self._query_template.expand(query=query)
        for i, result in grouped(resource.get()):
            yield self.graph.hydrate(assembled(result))

    def _query_with_score(self, query, order):
        resource = self._query_template.expand(query=query, order=order)
        for i, result in grouped(resource.get()):
            meta = assembled(result)
            yield self.graph.hydrate(meta), meta["score"]

    def query_by_index(self, query):
        return self._query_with_score(query, "index")

    def query_by_relevance(self, query):
        return self._query_with_score(query, "relevance")

    def query_by_score(self, query):
        return self._query_with_score(query, "score")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


""" HTTPStream is an HTTP client library for Python that wraps the
standard library HTTP client with a convenient resource-based interface
and also provides support for incremental JSON document retrieval and
RFC 6570 URI Templates.
"""


__author__ = "Nigel Small"
__copyright__ = "2013-2014, Nigel Small"
__email__ = "nigel@nigelsmall.com"
__license__ = "Apache License, Version 2.0"
__version__ = "1.5.0"

from .http import *
from .watch import watch


def head(uri, if_modified_since=None, headers=None, redirect_limit=5, **kwargs):
    """ Issue an HTTP ``HEAD`` request to a given `uri`.

    :param uri: target URI for the request
    :param headers: dictionary of extra headers to send (optional)
    :param redirect_limit: maximum number of redirects to follow (optional, default=5)
    :param kwargs: see :func:`get <httpstream.get>` for other keyword arguments
    :return: file-like :class:`Response <httpstream.Response>` object from which
        content can be read
    """
    return Resource(uri).head(if_modified_since, headers, redirect_limit, **kwargs)


def get(uri, if_modified_since=None, headers=None, redirect_limit=5, **kwargs):
    """ Issue an HTTP ``GET`` request to a given `uri`.

    :param uri: target URI for the request
    :param headers: dictionary of extra headers to send (optional)
    :param redirect_limit: maximum number of redirects to follow (optional, default=5)
    :param product: name or (name, version) tuple to be passed in the ``User-Agent``
        header (optional)
    :param chunk_size: number of bytes to retrieve per chunk (optional, default=4096)
    :param cache: boolean flag to allow caching so response content can be stored
        for multiple reads (optional)
    :return: file-like :class:`Response <httpstream.Response>` object from which
        content can be read
    """
    return Resource(uri).get(if_modified_since, headers, redirect_limit, **kwargs)


def put(uri, body=None, headers=None, **kwargs):
    """ Issue an HTTP ``PUT`` request to a given `uri`, optionally with a payload.

    :param uri: target URI for the request
    :param body: payload to be sent with the request (optional)
    :param headers: dictionary of extra headers to send (optional)
    :param kwargs: see :func:`get <httpstream.get>` for other keyword arguments
    :return: file-like :class:`Response <httpstream.Response>` object from which
        content can be read
    """
    return Resource(uri).put(body, headers, **kwargs)


def patch(uri, body=None, headers=None, **kwargs):
    """ Issue an HTTP ``PUT`` request to a given `uri`, optionally with a payload.

    :param uri: target URI for the request
    :param body: payload to be sent with the request (optional)
    :param headers: dictionary of extra headers to send (optional)
    :param kwargs: see :func:`get <httpstream.get>` for other keyword arguments
    :return: file-like :class:`Response <httpstream.Response>` object from which
        content can be read
    """
    return Resource(uri).patch(body, headers, **kwargs)


def post(uri, body=None, headers=None, **kwargs):
    """ Issue an HTTP ``POST`` request to a given `uri`, optionally with a payload.

    :param uri: target URI for the request
    :param body: payload to be sent with the request (optional)
    :param headers: dictionary of extra headers to send (optional)
    :param kwargs: see :func:`get <httpstream.get>` for other keyword arguments
    :return: file-like :class:`Response <httpstream.Response>` object from which
        content can be read
    """
    return Resource(uri).post(body, headers, **kwargs)


def delete(uri, headers=None, **kwargs):
    """ Issue an HTTP ``DELETE`` request to a given `uri`.

    :param uri: target URI for the request
    :param headers: dictionary of extra headers to send (optional)
    :param kwargs: see :func:`get <httpstream.get>` for other keyword arguments
    :return: file-like :class:`Response <httpstream.Response>` object from which
        content can be read
    """
    return Resource(uri).delete(headers, **kwargs)


def download(uri, filename=None, headers=None, redirect_limit=5, **kwargs):
    """ Get a remote resource and save the content to a local file. If a local
    file already exists at the download destination, the modification time of
    that file is sent in the request headers as the `If-Modified-Since` value
    to perform a conditional GET. If the remote file has not been modified, a
    304 response should then be returned.

    :param uri:
    :param filename:
    :param headers:
    :param redirect_limit:
    :param kwargs:
    :return:
    """
    from datetime import datetime
    from os import utime
    from os.path import getmtime
    from time import time
    from .tardis import datetime_to_timestamp
    try:
        last_modified = datetime.fromtimestamp(getmtime(filename))
    except OSError:
        last_modified = None
    with get(uri, if_modified_since=last_modified, headers=headers, redirect_limit=redirect_limit, **kwargs) as source:
        if source.status_code == 200:
            if not filename:
                filename = source.filename
            with open(filename, "wb") as destination:
                finished = False
                while not finished:
                    data = source.read(8192)
                    if data:
                        destination.write(data)
                    else:
                        finished = True
            utime(filename, (time(), datetime_to_timestamp(source.last_modified)))
            return True
        elif source.status_code == 304:
            return False
        else:
            raise ValueError("Unexpected status code %s from download response" % source.status_code)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function

import sys

from . import get, download


if __name__ == "__main__":
    command, args = sys.argv[0], sys.argv[1:]
    if len(args) == 1:
        sys.stdout.write(get(args[0]).content)
    elif len(args) == 2:
        download(args[0], args[1])
    else:
        print("Usage: foo bar")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


""" Deprecated features for core module.
"""


from py2neo.core import Graph, PropertyContainer, Node, Rel, Path, Relationship
from py2neo.error import BindError, GraphError
from py2neo.packages.httpstream.numbers import BAD_REQUEST
from py2neo.util import deprecated, flatten, ustr


__all__ = []


@deprecated("Use `pull` instead")
def _graph_get_properties(self, *entities):
    self.pull(*entities)
    return [entity.properties for entity in entities]


Graph.get_properties = _graph_get_properties


@deprecated("Use `properties` attribute instead")
def _property_container_get_cached_properties(self):
    return self.properties

@deprecated("Use `pull` method on `properties` attribute instead")
def _property_container_get_properties(self):
    if self.bound:
        self.properties.pull()
    return self.properties

@deprecated("Use `push` method on `properties` attribute instead")
def _property_container_set_properties(self, properties):
    self.properties.replace(properties)
    if self.bound:
        self.properties.push()

@deprecated("Use `push` method on `properties` attribute instead")
def _property_container_delete_properties(self):
    self.properties.clear()
    try:
        self.properties.push()
    except BindError:
        pass


PropertyContainer.get_cached_properties = _property_container_get_cached_properties
PropertyContainer.get_properties = _property_container_get_properties
PropertyContainer.set_properties = _property_container_set_properties
PropertyContainer.delete_properties = _property_container_delete_properties


@deprecated("Use `add` or `update` method of `labels` property instead")
def _node_add_labels(self, *labels):
    labels = [ustr(label) for label in set(flatten(labels))]
    self.labels.update(labels)
    try:
        self.labels.push()
    except GraphError as err:
        if err.response.status_code == BAD_REQUEST and err.cause.exception == 'ConstraintViolationException':
            raise ValueError(err.cause.message)
        else:
            raise

@deprecated("Use graph.create(Path(node, ...)) instead")
def _node_create_path(self, *items):
    path = Path(self, *items)
    return _path_create(path, self.graph)

@deprecated("Use Graph.delete instead")
def _node_delete(self):
    self.graph.delete(self)

@deprecated("Use Cypher query instead")
def _node_delete_related(self):
    from py2neo.cypher.util import StartOrMatch
    graph = self.graph
    foreach_symbol = "|" if graph.supports_foreach_pipe else ":"
    statement = (
        StartOrMatch(graph).node("a", "{a}").string +
        "MATCH (a)-[rels*0..]-(z) "
        "FOREACH(r IN rels %s DELETE r) "
        "DELETE a, z" % foreach_symbol
    )
    graph.cypher.post(statement, {"a": self._id})

@deprecated("Use `labels` property instead")
def _node_get_labels(self):
    self.labels.pull()
    return self.labels

@deprecated("Use graph.create_unique(Path(node, ...)) instead")
def _node_get_or_create_path(self, *items):
    path = Path(self, *items)
    return _path_get_or_create(path, self.graph)

@deprecated("Use Cypher query instead")
def _node_isolate(self):
    from py2neo.cypher.util import StartOrMatch
    graph = self.graph
    query = StartOrMatch(graph).node("a", "{a}").string + "MATCH a-[r]-b DELETE r"
    graph.cypher.post(query, {"a": self._id})

@deprecated("Use `remove` method of `labels` property instead")
def _node_remove_labels(self, *labels):
    from py2neo.batch import WriteBatch
    labels = [ustr(label) for label in set(flatten(labels))]
    batch = WriteBatch(self.graph)
    for label in labels:
        batch.remove_label(self, label)
    batch.run()

@deprecated("Use `clear` and `update` methods of `labels` property instead")
def _node_set_labels(self, *labels):
    labels = [ustr(label) for label in set(flatten(labels))]
    self.labels.clear()
    self.labels.add(*labels)


Node.add_labels = _node_add_labels
Node.create_path = _node_create_path
Node.delete = _node_delete
Node.delete_related = _node_delete_related
Node.get_labels = _node_get_labels
Node.get_or_create_path = _node_get_or_create_path
Node.isolate = _node_isolate
Node.remove_labels = _node_remove_labels
Node.set_labels = _node_set_labels


@deprecated("Use graph.delete instead")
def _rel_delete(self):
    self.resource.delete()


Rel.delete = _rel_delete


def _path__create_query(self, graph, unique):
    from py2neo.cypher.util import StartOrMatch
    start_or_match_clause = StartOrMatch(graph)
    path, values, params = [], [], {}

    def append_node(i, node):
        if node is None:
            path.append("(n{0})".format(i))
            values.append("n{0}".format(i))
        elif node.bound:
            path.append("(n{0})".format(i))
            start_or_match_clause.node("n%s" % i, "{i%s}" % i)
            params["i{0}".format(i)] = node._id
            values.append("n{0}".format(i))
        else:
            path.append("(n{0} {{p{0}}})".format(i))
            params["p{0}".format(i)] = node.properties
            values.append("n{0}".format(i))

    def append_rel(i, rel):
        if rel.properties:
            path.append("-[r{0}:`{1}` {{q{0}}}]->".format(i, rel.type))
            params["q{0}".format(i)] = rel.properties
            values.append("r{0}".format(i))
        else:
            path.append("-[r{0}:`{1}`]->".format(i, rel.type))
            values.append("r{0}".format(i))

    append_node(0, self._Path__nodes[0])
    for i, rel in enumerate(self._Path__rels):
        append_rel(i, rel)
        append_node(i + 1, self._Path__nodes[i + 1])
    clauses = []
    if len(start_or_match_clause) > 0:
        clauses.append(start_or_match_clause.string)
    if unique:
        clauses.append("CREATE UNIQUE p={0}".format("".join(path)))
    else:
        clauses.append("CREATE p={0}".format("".join(path)))
    clauses.append("RETURN p")
    query = "\n".join(clauses)
    return query, params


def _path__create(self, graph, unique):
    query, params = _path__create_query(self, graph, unique=unique)
    try:
        results = graph.cypher.execute(query, params)
    except GraphError:
        raise NotImplementedError(
            "The Neo4j server at <{0}> does not support "
            "Cypher CREATE UNIQUE clauses or the query contains "
            "an unsupported property type".format(graph.uri)
        )
    else:
        for row in results:
            return row.p

@deprecated("Use Graph.create(Path(...)) instead")
def _path_create(self, graph):
    return _path__create(self, graph, unique=False)

@deprecated("Use Graph.create_unique(Path(...)) instead")
def _path_get_or_create(self, graph):
    return _path__create(self, graph, unique=True)


Path.create = _path_create
Path.get_or_create = _path_get_or_create


@deprecated("Use Graph.delete instead")
def _relationship_delete(self):
    self.graph.delete(self)

@deprecated("Use `push` method on `properties` attribute instead")
def _relationship_delete_properties(self):
    self.properties.clear()
    try:
        self.properties.push()
    except BindError:
        pass

@deprecated("Use `properties` attribute instead")
def _relationship_get_cached_properties(self):
    return self.properties

@deprecated("Use `pull` method on `properties` attribute instead")
def _relationship_get_properties(self):
    if self.bound:
        self.properties.pull()
    return self.properties

@deprecated("Use `push` method on `properties` attribute instead")
def _relationship_set_properties(self, properties):
    self.properties.replace(properties)
    if self.bound:
        self.properties.push()

@deprecated("Use properties.update and push instead")
def _relationship_update_properties(self, properties):
    if self.bound:
        from py2neo.cypher.lang import cypher_escape
        from py2neo.cypher.util import StartOrMatch
        graph = self.graph
        query, params = [StartOrMatch(graph).node("a", "{A}").string], {"A": self._id}
        for i, (key, value) in enumerate(properties.items()):
            value_tag = "V" + str(i)
            query.append("SET a.%s={" + value_tag + "}" % cypher_escape(key))
            params[value_tag] = value
        query.append("RETURN a")
        rel = graph.cypher.execute_one("\n".join(query), params)
        self._properties = rel.__metadata__["data"]
    else:
        self._properties.update(properties)


Relationship.delete = _relationship_delete
Relationship.delete_properties = _relationship_delete_properties
Relationship.get_cached_properties = _relationship_get_cached_properties
Relationship.get_properties = _relationship_get_properties
Relationship.set_properties = _relationship_set_properties
Relationship.update_properties = _relationship_update_properties



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2012-2014 Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Incremental JSON parser.
"""


from __future__ import unicode_literals

try:
    from builtins import chr as _chr
except ImportError:
    from __builtin__ import unichr as _chr
from itertools import groupby
from string import digits, whitespace


__all__ = ["JSONStream", "assembled", "grouped"]


class AwaitingData(BaseException):
    """ Raised when data is temporarily unavailable.
    """

    def __init__(self, *args, **kwargs):
        super(AwaitingData, self).__init__(*args, **kwargs)


class EndOfStream(BaseException):
    """ Raised when stream is exhausted.
    """

    def __init__(self, *args, **kwargs):
        super(EndOfStream, self).__init__(*args, **kwargs)


class UnexpectedCharacter(ValueError):
    """ Raised when a unexpected character is encountered.
    """

    def __init__(self, *args, **kwargs):
        super(UnexpectedCharacter, self).__init__(*args, **kwargs)


class TextStream:

    def __init__(self):
        self.__data = []
        self.__current_line = 0
        self.__current_char = 0
        self.__writable = True
        self.__marked_line = 0
        self.__marked_char = 0

    def close(self):
        self.__writable = False

    def peek(self):
        if self.__current_line < len(self.__data):
            line = self.__data[self.__current_line]
            if self.__current_char < len(line):
                return line[self.__current_char]
            else:
                # no more characters on this line, jump to the next
                self.__current_line += 1
                self.__current_char = 0
                if self.__current_line < len(self.__data):
                    return self.__data[self.__current_line][self.__current_char]
        if self.__writable:
            raise AwaitingData()
        else:
            raise EndOfStream()

    def read(self):
        if self.__current_line < len(self.__data):
            line = self.__data[self.__current_line]
            if self.__current_char < len(line):
                ch = line[self.__current_char]
                self.__current_char += 1
                return ch
            else:
                self.__current_line += 1
                if self.__current_line < len(self.__data):
                    self.__current_char = 1
                    return self.__data[self.__current_line][0]
                else:
                    self.__current_char = 0
        if self.__writable:
            raise AwaitingData()
        else:
            raise EndOfStream()

    def read_any(self, allowed):
        out = []
        start = self.__current_char
        while True:
            if self.__current_line < len(self.__data):
                line = self.__data[self.__current_line]
                if self.__current_char < len(line):
                    ch = self.__data[self.__current_line][self.__current_char]
                    if ch in allowed:
                        # move forward
                        self.__current_char += 1
                    else:
                        # return everything between start and here
                        out.append(line[start:self.__current_char])
                        return "".join(out)
                else:
                    # no more characters on this line
                    out.append(line[start:])
                    self.__current_line += 1
                    self.__current_char = 0
                    start = 0
            elif self.__writable:
                raise AwaitingData()
            else:
                return "".join(out)

    #def read_until(self, marker):
    #    out = []
    #    line = self.__current_line
    #    start = self.__current_char
    #    while True:
    #        try:
    #            end = self.__data[line].index(marker, start)
    #        except IndexError:  # no more lines
    #            if self.__writable:
    #                raise AwaitingData()
    #            else:
    #                raise EndOfStream()
    #        except ValueError:  # not found
    #            out.append(self.__data[line][start:])
    #            line += 1
    #            start = 0
    #        else:
    #            # found
    #            self.__current_line = line
    #            self.__current_char = end + 1
    #            out.append(self.__data[line][start:self.__current_char])
    #            return "".join(out)

    def read_until_any(self, markers):
        out = []
        line = self.__current_line
        start = self.__current_char
        while True:
            if line < len(self.__data):
                try:
                    end = min(pos
                              for pos in [self.__data[line].find(marker, start)
                                          for marker in markers]
                              if pos >= 0)
                except ValueError:  # not found
                    out.append(self.__data[line][start:])
                    line += 1
                    start = 0
                else:
                    # found
                    self.__current_line = line
                    self.__current_char = end + 1
                    out.append(self.__data[line][start:self.__current_char])
                    return "".join(out)
            elif self.__writable:
                raise AwaitingData()
            else:
                raise EndOfStream()

    def peek_after_any(self, markers):
        """
        skips any characters in the marker set and returns a peek of the next
        """
        while True:
            if self.__current_line < len(self.__data):
                line = self.__data[self.__current_line]
                if self.__current_char < len(line):
                    ch = self.__data[self.__current_line][self.__current_char]
                    if ch in markers:
                        # skip
                        self.__current_char += 1
                    else:
                        # peek
                        return ch
                else:
                    # no more characters on this line
                    self.__current_line += 1
                    self.__current_char = 0
            elif self.__writable:
                raise AwaitingData()
            else:
                raise EndOfStream()

    def write(self, data):
        if not self.__writable:
            raise IOError("Stream is not writable")
        if data:
            # so we can guarantee no line is empty
            self.__data.append(data)

    def mark(self):
        self.__marked_line = self.__current_line
        self.__marked_char = self.__current_char

    def undo(self):
        self.__current_line = self.__marked_line
        self.__current_char = self.__marked_char


class Tokeniser(object):

    def __init__(self):
        self.__text = TextStream()

    def close(self):
        self.__text.close()

    def write(self, data):
        """Write raw JSON data to the decoder stream.
        """
        self.__text.write(data)

    def _read_literal(self, literal):
        self.__text.mark()
        try:
            for expected in literal:
                actual = self.__text.read()
                if actual != expected:
                    raise UnexpectedCharacter(actual)
        except AwaitingData:
            self.__text.undo()
            raise
        return literal

    def _read_string(self):
        self.__text.mark()
        try:
            src, value = [self._read_literal('"')], []
            while True:
                chunk = self.__text.read_until_any(('"', '\\'))
                src.append(chunk)
                value.append(chunk[:-1])
                if chunk.endswith('\\'):
                    ch = self.__text.read()
                    src.append(ch)
                    if ch in '"/\\':
                        value.append(ch)
                    elif ch == 'b':
                        value.append('\b')
                    elif ch == 'f':
                        value.append('\f')
                    elif ch == 'n':
                        value.append('\n')
                    elif ch == 'r':
                        value.append('\r')
                    elif ch == 't':
                        value.append('\t')
                    elif ch == 'u':
                        n = 0
                        for i in range(4):
                            ch = self.__text.read()
                            src.append(ch)
                            n = 16 * n + int(ch, 16)
                        value.append(_chr(n))
                    else:
                        raise UnexpectedCharacter(ch)
                else:
                    return "".join(src), "".join(value)
        except AwaitingData:
            self.__text.undo()
            raise

    def _read_number(self):
        src = []
        has_fractional_part = False
        has_exponent = False
        self.__text.mark()
        try:
            # check for sign
            ch = self.__text.read_any("-")
            if ch:
                src.append(ch)
            # read integer part
            src.append(self.__text.read_any(digits))
            # read fractional part
            ch = self.__text.read_any(".")
            if ch:
                has_fractional_part = True
                src.append(ch)
                src.append(self.__text.read_any(digits))
            # read exponent
            ch = self.__text.read_any('Ee')
            if ch:
                has_exponent = True
                src.append(ch)
                ch = self.__text.read_any('+-')
                if ch:
                    src.append(ch)
                src.append(self.__text.read_any(digits))
        except AwaitingData:
            # number potentially incomplete: need to wait for
            # further data or end of stream
            self.__text.undo()
            raise
        str_src = "".join(src)
        if has_fractional_part or has_exponent:
            return str_src, float(str_src)
        else:
            return str_src, int(str_src)

    def read_token(self):
        """ Read token

        """
        ch = self.__text.peek_after_any(whitespace)
        if ch in ',:[]{}':
            return self.__text.read(), None
        if ch == '"':
            return self._read_string()
        if ch in '0123456789+-':
            return self._read_number()
        if ch == 't':
            return self._read_literal("true"), True
        if ch == 'f':
            return self._read_literal("false"), False
        if ch == 'n':
            return self._read_literal("null"), None
        raise UnexpectedCharacter(ch)


# Token constants used for expectation management
VALUE = 0x01
OPEN_BRACKET = 0x02
CLOSE_BRACKET = 0x04
OPEN_BRACE = 0x08
CLOSE_BRACE = 0x10
COMMA = 0x20
COLON = 0x40

VALUE_OR_OPEN = VALUE | OPEN_BRACKET | OPEN_BRACE
VALUE_BRACKET_OR_OPEN_BRACE = VALUE | OPEN_BRACKET | CLOSE_BRACKET | OPEN_BRACE
COMMA_OR_CLOSE_BRACKET = COMMA | CLOSE_BRACKET
COMMA_OR_CLOSE_BRACE = COMMA | CLOSE_BRACE
VALUE_OR_CLOSE_BRACE = VALUE | CLOSE_BRACE


class JSONStream(object):
    """ Streaming JSON decoder. This class both expects Unicode input and will
    produce Unicode output.
    """

    def __init__(self, source):
        self.tokeniser = Tokeniser()
        self.source = iter(source)
        self.path = []
        self._expectation = VALUE_OR_OPEN

    def _in_array(self):
        return self.path and isinstance(self.path[-1], int)

    def _in_object(self):
        return self.path and not isinstance(self.path[-1], int)

    def __iter__(self):
        while True:
            try:
                try:
                    self.tokeniser.write(next(self.source))
                except StopIteration:
                    self.tokeniser.close()
                while True:
                    try:
                        src, value = self.tokeniser.read_token()
                        if src == ',':
                            if not self._expectation & COMMA:
                                raise UnexpectedCharacter(",")
                            self._expectation = VALUE_OR_OPEN
                        elif src == ':':
                            if not self._expectation & COLON:
                                raise UnexpectedCharacter(":")
                            self._expectation = VALUE_OR_OPEN
                        elif src == '[':
                            yield tuple(self.path), []
                            if not self._expectation & OPEN_BRACKET:
                                raise UnexpectedCharacter("[")
                            self.path.append(0)
                            self._expectation = VALUE_BRACKET_OR_OPEN_BRACE
                        elif src == ']':
                            if not self._expectation & CLOSE_BRACKET:
                                raise UnexpectedCharacter("]")
                            self.path.pop()
                            if self._in_array():
                                self.path[-1] += 1
                                self._expectation = COMMA_OR_CLOSE_BRACKET
                            elif self._in_object():
                                self.path[-1] = None
                                self._expectation = COMMA_OR_CLOSE_BRACE
                            else:
                                self._expectation = VALUE_OR_OPEN
                        elif src == '{':
                            yield tuple(self.path), {}
                            if not self._expectation & OPEN_BRACE:
                                raise UnexpectedCharacter("{")
                            self.path.append(None)
                            self._expectation = VALUE_OR_CLOSE_BRACE
                        elif src == '}':
                            if not self._expectation & CLOSE_BRACE:
                                raise UnexpectedCharacter("}")
                            self.path.pop()
                            if self._in_array():
                                self.path[-1] += 1
                                self._expectation = COMMA_OR_CLOSE_BRACKET
                            elif self._in_object():
                                self.path[-1] = None
                                self._expectation = COMMA_OR_CLOSE_BRACE
                            else:
                                self._expectation = VALUE_OR_OPEN
                        else:
                            if not self._expectation & VALUE:
                                raise UnexpectedCharacter(src)
                            if self._in_array():
                                # array value
                                yield tuple(self.path), value
                                self.path[-1] += 1
                                self._expectation = COMMA_OR_CLOSE_BRACKET
                            elif self._in_object():
                                if self.path[-1] is None:
                                    # object key (no yield)
                                    self.path[-1] = value
                                    self._expectation = COLON
                                else:
                                    # object value
                                    yield tuple(self.path), value
                                    self.path[-1] = None
                                    self._expectation = COMMA_OR_CLOSE_BRACE
                            else:
                                # simple value
                                yield tuple(self.path), value
                    except AwaitingData:
                        break
            except EndOfStream:
                break


def _merged(obj, key, value):
    """ Returns object with value merged at a position described by iterable
    key. The key describes a navigable path through the object hierarchy with
    integer items describing list indexes and other types of items describing
    dictionary keys.

        >>> obj = None
        >>> obj = _merged(obj, ("drink",), "lemonade")
        >>> obj
        {'drink': 'lemonade'}
        >>> obj = _merged(obj, ("cutlery", 0), "knife")
        >>> obj = _merged(obj, ("cutlery", 1), "fork")
        >>> obj = _merged(obj, ("cutlery", 2), "spoon")
        >>> obj
        {'cutlery': ['knife', 'fork', 'spoon'], 'drink': 'lemonade'}

    """
    if key:
        k = key[0]
        if isinstance(k, int):
            if isinstance(obj, list):
                obj = list(obj)
            else:
                obj = []
            while len(obj) <= k:
                obj.append(None)
        else:
            if isinstance(obj, dict):
                obj = dict(obj)
            else:
                obj = {}
            obj.setdefault(k, None)
        obj[k] = _merged(obj[k], key[1:], value)
        return obj
    else:
        return value


def assembled(iterable):
    """ Returns a JSON-derived value from a set of key-value pairs as produced
    by the JSONStream process. This operates in a similar way to the built-in
    `dict` function. Internally, this uses the `merged` function on each pair
    to build the return value.

        >>> data = [
        ...     (("drink",), "lemonade"),
        ...     (("cutlery", 0), "knife"),
        ...     (("cutlery", 1), "fork"),
        ...     (("cutlery", 2), "spoon"),
        ... ]
        >>> assembled(data)
        {'cutlery': ['knife', 'fork', 'spoon'], 'drink': 'lemonade'}

    :param iterable: key-value pairs to be merged into assembled value
    """
    obj = None
    for key, value in iterable:
        obj = _merged(obj, key, value)
    return obj


def _group(iterable, level):
    for key, value in iterable:
        yield key[level:], value


def grouped(iterable, level=1):
    def _group_key(item):
        key, value = item
        if len(key) >= level:
            return key[0:level]
        else:
            return None
    for key, value in groupby(iterable, _group_key):
        if key is not None:
            yield key, _group(value, level)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from importlib import import_module
import json

from py2neo.error import GraphError


class CypherError(GraphError):

    default_message = "The server returned a Cypher error"
    status_code_map = {}

    statement = None
    parameters = None

    def __init__(self, message, **kwargs):
        GraphError.__init__(self, message, **kwargs)
        if self.request:
            request_body = json.loads(self.request.body)
            self.statement = request_body.get("query")
            self.parameters = request_body.get("params")


class CypherTransactionError(CypherError):
    """ 
    """

    code = None
    message = None

    @classmethod
    def hydrate(cls, data):
        code = data["code"]
        message = data["message"]
        _, classification, category, title = code.split(".")
        error_module = import_module("py2neo.cypher.error." + category.lower())
        error_cls = getattr(error_module, title)
        inst = error_cls(message)
        inst.code = code
        inst.message = message
        return inst


class ClientError(CypherTransactionError):
    """ The Client sent a bad request - changing the request might yield a successful outcome.
    """


class DatabaseError(CypherTransactionError):
    """ The database failed to service the request.
    """


class TransientError(CypherTransactionError):
    """ The database cannot service the request right now, retrying later might yield a successful outcome.
    """



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.cypher.error.core import ClientError, DatabaseError


class ReadOnly(ClientError):
    """ This is a read only database, writing or modifying the database
    is not allowed.
    """


class CorruptSchemaRule(DatabaseError):
    """ A malformed schema rule was encountered. Please contact your
    support representative.
    """


class FailedIndex(DatabaseError):
    """ The request (directly or indirectly) referred to an index that
    is in a failed state. The index needs to be dropped and recreated
    manually.
    """


class UnknownFailure(DatabaseError):
    """ An unknown failure occurred.
    """



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.error import GraphError


class StartOrMatch(object):

    def __init__(self, graph):
        if graph is None:
            raise GraphError("All auto-generated Cypher statements require a Graph instance")
        self.graph = graph
        self.nodes = []
        self.relationships = []

    def __len__(self):
        return len(self.nodes) + len(self.relationships)

    def node(self, name, selector):
        self.nodes.append((name, selector))
        return self

    def relationship(self, name, selector):
        self.relationships.append((name, selector))
        return self

    @property
    def string(self):
        if self.nodes or self.relationships:
            if self.graph.supports_start_clause:
                s = []
                for name, selector in self.nodes:
                    s.append("%s=node(%s)" % (name, selector))
                for name, selector in self.relationships:
                    s.append("%s=rel(%s)" % (name, selector))
                return "START " + ",".join(s) + "\n"
            else:
                s = []
                for name, selector in self.nodes:
                    if selector == "*":
                        s.append("MATCH (%s)\n" % name)
                    else:
                        s.append("MATCH (%s) WHERE id(%s)=%s\n" % (name, name, selector))
                for name, selector in self.relationships:
                    if selector == "*":
                        s.append("MATCH ()-[%s]->()\n" % name)
                    else:
                        s.append("MATCH ()-[%s]->() WHERE id(%s)=%s\n" % (name, name, selector))
                return "".join(s)
        else:
            return ""



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.batch import ReadBatch, WriteBatch, Job, Target
from py2neo.core import Node, Relationship, NodePointer
from py2neo.packages.httpstream.packages.urimagic import percent_encode


__all__ = ["LegacyReadBatch", "LegacyWriteBatch"]


class LegacyReadBatch(ReadBatch):
    """ Generic batch execution facility for data read requests,
    """

    def append_get(self, uri):
        return self.append(Job("GET", Target(uri)))

    def _index(self, content_type, index):
        """ Fetch an Index object.
        """
        from py2neo.legacy.index import Index
        if isinstance(index, Index):
            if content_type == index._content_type:
                return index
            else:
                raise TypeError("Index is not for {0}s".format(content_type))
        else:
            return self.graph.legacy.get_or_create_index(content_type, str(index))

    def get_indexed_nodes(self, index, key, value):
        """ Fetch all nodes indexed under a given key-value pair.

        :param index: index name or instance
        :type index: :py:class:`str` or :py:class:`Index`
        :param key: key under which nodes are indexed
        :type key: :py:class:`str`
        :param value: value under which nodes are indexed
        :return: batch request object
        """
        index = self._index(Node, index)
        uri = index._searcher_stem_for_key(key) + percent_encode(value)
        return self.append_get(uri)


class LegacyWriteBatch(WriteBatch):
    """ Generic batch execution facility for data write requests. Most methods
    return a :py:class:`BatchRequest <py2neo.neo4j.BatchRequest>` object that
    can be used as a reference in other methods. See the
    :py:meth:`create <py2neo.neo4j.WriteBatch.create>` method for an example
    of this.
    """

    def append_post(self, uri, body=None):
        return self.append(Job("POST", Target(uri), body))

    def append_delete(self, uri):
        return self.append(Job("DELETE", Target(uri)))

    def _uri_for(self, resource, *segments, **kwargs):
        """ Return a relative URI in string format for the entity specified
        plus extra path segments.
        """
        if isinstance(resource, int):
            uri = "{{{0}}}".format(resource)
        elif isinstance(resource, NodePointer):
            uri = "{{{0}}}".format(resource.address)
        elif isinstance(resource, Job):
            uri = "{{{0}}}".format(self.find(resource))
        else:
            uri = resource.ref
        if segments:
            if not uri.endswith("/"):
                uri += "/"
            uri += "/".join(map(percent_encode, segments))
        query = kwargs.get("query")
        if query is not None:
            uri += "?" + query
        return uri

    def _index(self, content_type, index):
        """ Fetch an Index object.
        """
        from py2neo.legacy.index import Index
        if isinstance(index, Index):
            if content_type == index._content_type:
                return index
            else:
                raise TypeError("Index is not for {0}s".format(content_type))
        else:
            return self.graph.legacy.get_or_create_index(content_type, str(index))

    def __init__(self, graph):
        super(LegacyWriteBatch, self).__init__(graph)
        self.__new_uniqueness_modes = None

    @property
    def supports_index_uniqueness_modes(self):
        return self.graph.legacy.supports_index_uniqueness_modes

    def _assert_can_create_or_fail(self):
        if not self.supports_index_uniqueness_modes:
            raise NotImplementedError("Uniqueness mode `create_or_fail` "
                                      "requires version 1.9 or above")

    ### ADD TO INDEX ###

    def _add_to_index(self, cls, index, key, value, entity, query=None):
        uri = self._uri_for(self._index(cls, index), query=query)
        return self.append_post(uri, {
            "key": key,
            "value": value,
            "uri": self._uri_for(entity),
        })

    def add_to_index(self, cls, index, key, value, entity):
        """ Add an existing node or relationship to an index.

        :param cls: the type of indexed entity
        :type cls: :py:class:`Node <py2neo.neo4j.Node>` or
                   :py:class:`Relationship <py2neo.neo4j.Relationship>`
        :param index: index or index name
        :type index: :py:class:`Index <py2neo.neo4j.Index>` or :py:class:`str`
        :param key: index entry key
        :type key: :py:class:`str`
        :param value: index entry value
        :param entity: node or relationship to add to the index
        :type entity: concrete or reference
        :return: batch request object
        """
        return self._add_to_index(cls, index, key, value, entity)

    def add_to_index_or_fail(self, cls, index, key, value, entity):
        """ Add an existing node or relationship uniquely to an index, failing
        the entire batch if such an entry already exists.

        .. warning::
            Uniqueness modes for legacy indexes have been broken in recent
            server versions and therefore this method may not work as expected.

        :param cls: the type of indexed entity
        :type cls: :py:class:`Node <py2neo.neo4j.Node>` or
                   :py:class:`Relationship <py2neo.neo4j.Relationship>`
        :param index: index or index name
        :type index: :py:class:`Index <py2neo.neo4j.Index>` or :py:class:`str`
        :param key: index entry key
        :type key: :py:class:`str`
        :param value: index entry value
        :param entity: node or relationship to add to the index
        :type entity: concrete or reference
        :return: batch request object
        """
        self._assert_can_create_or_fail()
        query = "uniqueness=create_or_fail"
        return self._add_to_index(cls, index, key, value, entity, query)

    def get_or_add_to_index(self, cls, index, key, value, entity):
        """ Fetch a uniquely indexed node or relationship if one exists,
        otherwise add an existing entity to the index.

        :param cls: the type of indexed entity
        :type cls: :py:class:`Node <py2neo.neo4j.Node>` or
                   :py:class:`Relationship <py2neo.neo4j.Relationship>`
        :param index: index or index name
        :type index: :py:class:`Index <py2neo.neo4j.Index>` or :py:class:`str`
        :param key: index entry key
        :type key: :py:class:`str`
        :param value: index entry value
        :param entity: node or relationship to add to the index
        :type entity: concrete or reference
        :return: batch request object
        """
        if self.supports_index_uniqueness_modes:
            query = "uniqueness=get_or_create"
        else:
            query = "unique"
        return self._add_to_index(cls, index, key, value, entity, query)

    ### CREATE IN INDEX ###

    def _create_in_index(self, cls, index, key, value, abstract, query=None):
        uri = self._uri_for(self._index(cls, index), query=query)
        if cls is Node:
            node = Node.cast(abstract)
            return self.append_post(uri, {
                "key": key,
                "value": value,
                "properties": node.properties,
            })
        elif cls is Relationship:
            relationship = Relationship.cast(abstract)
            return self.append_post(uri, {
                "key": key,
                "value": value,
                "start": self._uri_for(abstract.start_node),
                "type": str(abstract.type),
                "end": self._uri_for(abstract.end_node),
                "properties": relationship.properties,
            })
        else:
            raise TypeError(cls)

    # Removed create_in_index as parameter combination not supported by server

    def create_in_index_or_fail(self, cls, index, key, value, abstract=None):
        """ Create a new node or relationship and add it uniquely to an index,
        failing the entire batch if such an entry already exists.

        .. warning::
            Uniqueness modes for legacy indexes have been broken in recent
            server versions and therefore this method may not work as expected.

        :param cls: the type of indexed entity
        :type cls: :py:class:`Node <py2neo.neo4j.Node>` or
                   :py:class:`Relationship <py2neo.neo4j.Relationship>`
        :param index: index or index name
        :type index: :py:class:`Index <py2neo.neo4j.Index>` or :py:class:`str`
        :param key: index entry key
        :type key: :py:class:`str`
        :param value: index entry value
        :param abstract: abstract node or relationship to create
        :return: batch request object
        """
        self._assert_can_create_or_fail()
        query = "uniqueness=create_or_fail"
        return self._create_in_index(cls, index, key, value, abstract, query)

    def get_or_create_in_index(self, cls, index, key, value, abstract=None):
        """ Fetch a uniquely indexed node or relationship if one exists,
        otherwise create a new entity and add that to the index.

        :param cls: the type of indexed entity
        :type cls: :py:class:`Node <py2neo.neo4j.Node>` or
                   :py:class:`Relationship <py2neo.neo4j.Relationship>`
        :param index: index or index name
        :type index: :py:class:`Index <py2neo.neo4j.Index>` or :py:class:`str`
        :param key: index entry key
        :type key: :py:class:`str`
        :param value: index entry value
        :param abstract: abstract node or relationship to create
        :return: batch request object
        """
        if self.supports_index_uniqueness_modes:
            query = "uniqueness=get_or_create"
        else:
            query = "unique"
        return self._create_in_index(cls, index, key, value, cls.cast(abstract), query)

    ### REMOVE FROM INDEX ###

    def remove_from_index(self, cls, index, key=None, value=None, entity=None):
        """ Remove any nodes or relationships from an index that match a
        particular set of criteria. Allowed parameter combinations are:

        `key`, `value`, `entity`
            remove a specific node or relationship indexed under a given
            key-value pair

        `key`, `entity`
            remove a specific node or relationship indexed against a given key
            and with any value

        `entity`
            remove all occurrences of a specific node or relationship
            regardless of key or value

        :param cls: the type of indexed entity
        :type cls: :py:class:`Node <py2neo.neo4j.Node>` or
                   :py:class:`Relationship <py2neo.neo4j.Relationship>`
        :param index: index or index name
        :type index: :py:class:`Index <py2neo.neo4j.Index>` or :py:class:`str`
        :param key: index entry key
        :type key: :py:class:`str`
        :param value: index entry value
        :param entity: node or relationship to remove from the index
        :type entity: concrete or reference
        :return: batch request object
        """
        index = self._index(cls, index)
        if key and value and entity:
            uri = self._uri_for(index, key, value, entity._id)
        elif key and entity:
            uri = self._uri_for(index, key, entity._id)
        elif entity:
            uri = self._uri_for(index, entity._id)
        else:
            raise TypeError("Illegal parameter combination for index removal")
        return self.append_delete(uri)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.cypher.error.core import ClientError, DatabaseError


class ConstraintAlreadyExists(ClientError):
    """ Unable to perform operation because it would clash with a
    pre-existing constraint.
    """


class ConstraintVerificationFailure(ClientError):
    """ Unable to create constraint because data that exists in the
    database violates it.
    """


class ConstraintViolation(ClientError):
    """ A constraint imposed by the database was violated.
    """


class IllegalTokenName(ClientError):
    """ A token name, such as a label, relationship type or property
    key, used is not valid. Tokens cannot be empty strings and cannot
    be null.
    """


class IndexAlreadyExists(ClientError):
    """ Unable to perform operation because it would clash with a
    pre-existing index.
    """


class IndexBelongsToConstraint(ClientError):
    """ A requested operation can not be performed on the specified
    index because the index is part of a constraint. If you want to
    drop the index, for instance, you must drop the constraint.
    """


class LabelLimitReached(ClientError):
    """ The maximum number of labels supported has been reached, no
    more labels can be created.
    """


class NoSuchConstraint(ClientError):
    """ The request (directly or indirectly) referred to a constraint
    that does not exist.
    """


class NoSuchIndex(ClientError):
    """ The request (directly or indirectly) referred to an index that
    does not exist.
    """


class ConstraintCreationFailure(DatabaseError):
    """ Creating a requested constraint failed.
    """


class ConstraintDropFailure(DatabaseError):
    """ The database failed to drop a requested constraint.
    """


class IndexCreationFailure(DatabaseError):
    """ Failed to create an index.
    """


class IndexDropFailure(DatabaseError):
    """ The database failed to drop a requested index.
    """


class NoSuchLabel(DatabaseError):
    """ The request accessed a label that did not exist.
    """


class NoSuchPropertyKey(DatabaseError):
    """ The request accessed a property that does not exist.
    """


class NoSuchRelationshipType(DatabaseError):
    """ The request accessed a relationship type that does not exist.
    """


class NoSuchSchemaRule(DatabaseError):
    """ The request referred to a schema rule that does not exist.
    """



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. note::
   This module has been designed to work on Linux and may not operate
   correctly - or at all - on other platforms.

The ``store`` module provides a class for working with a Neo4j data store, typically contained
within the ``data/graph.db`` directory.
"""


import os
from shutil import copytree, rmtree


__all__ = ["GraphStore"]


class GraphStore(object):
    """ A physical database store on disk.
    """

    #: The full file path of this store.
    path = None

    @classmethod
    def for_server(cls, server):
        """ Return the store object for the given server.

        :arg server: A :class:`py2neo.server.GraphServer` object.
        :rtype: :class:`.GraphStore`

        """
        database_location = server.conf.get("neo4j-server", "org.neo4j.server.database.location")
        return GraphStore(os.path.join(server.home, database_location))

    def __init__(self, path):
        self.path = path

    def __repr__(self):
        return "<GraphStore path=%r>" % self.path

    @property
    def locked(self):
        """ Returns :const:`True` if store is currently in use,
        :const:`False` otherwise.
        """
        return os.path.isfile(os.path.join(self.path, "lock"))

    def drop(self, force=False):
        """ Delete this store directory.

        :param force:

        """
        if force or not self.locked:
            rmtree(self.path, ignore_errors=force)
        else:
            raise RuntimeError("Refusing to drop database store while in use")

    def load(self, path, force=False):
        if force or not self.locked:
            if not os.path.isdir(path):
                raise ValueError("Load source %r is not a directory" % path)
            rmtree(self.path, ignore_errors=force)
            copytree(path, self.path)
        else:
            raise RuntimeError("Refusing to load database store while in use")

    def save(self, path, force=False):
        if force or not self.locked:
            copytree(self.path, path)
        else:
            raise RuntimeError("Refusing to save database store while in use")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo import Node, Path, GraphError


__all__ = ["GregorianCalendar", "GregorianDate"]


class GregorianCalendar(object):
    """ A Gregorian calendar stored in a graph as a tree
    of ``(year)->(month)->(day)``.
    """

    #: The graph associated with this calendar.
    graph = None

    __instances = {}

    def __new__(cls, graph):
        if not graph.supports_node_labels:
            raise ValueError("Graph does not support node labels")
        try:
            inst = cls.__instances[graph]
        except KeyError:
            inst = super(GregorianCalendar, cls).__new__(cls)
            inst.graph = graph
            try:
                inst.graph.schema.create_uniqueness_constraint("Calendar", "name")
            except GraphError as error:
                if error.__class__.__name__ == "ConstraintViolationException":
                    pass
                else:
                    raise
            inst.root = inst.graph.merge_one("Calendar", "name", "Gregorian")
            cls.__instances[graph] = inst
        return inst

    def date(self, year, month=1, day=1):
        """ Pick a date from this calendar.

        :rtype: :class:`.GregorianDate`

        """
        return GregorianDate(self, year, month, day)


class GregorianDate(object):
    """ A date picked from a :class:`.GregorianCalendar`.
    """

    #: The calendar from which this date was picked.
    calendar = None

    #: The graph associated with this date.
    graph = None

    #: Full :class:`Path <py2neo.Path>` representing this date.
    path = None

    def __init__(self, calendar, year, month=1, day=1):
        self.calendar = calendar
        self.graph = self.calendar.graph
        self.path = Path(self.calendar.root,
                         "YEAR", Node("Year", key='%04d' % year, year=year),
                         "MONTH", Node("Month", key='%04d-%02d' % (year, month), year=year, month=month),
                         "DAY", Node("Day", key='%04d-%02d-%02d' % (year, month, day), year=year, month=month, day=day))
        self.graph.create_unique(self.path)

    @property
    def year(self):
        """ The year node for this date.

        :rtype: :class:`py2neo.Node`

        """
        return self.path.nodes[1]

    @property
    def month(self):
        """ The month node for this date.

        :rtype: :class:`py2neo.Node`

        """
        return self.path.nodes[2]

    @property
    def day(self):
        """ The day node for this date.

        :rtype: :class:`py2neo.Node`

        """
        return self.path.nodes[3]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json

from py2neo.core import Node, Rel, Rev, Path
from py2neo.lang import Writer
from py2neo.util import is_collection, ustr


__all__ = ["GeoffWriter"]


class GeoffWriter(Writer):
    """ Writer for Geoff data. This can be used to write to any
    file-like object, such as standard output::

        >>> from py2neo import Node
        >>> from py2neo.ext.geoff import GeoffWriter
        >>> import sys
        >>> writer = GeoffWriter(sys.stdout)
        >>> writer.write(Node("Person", name="Alice"))
        (139895901379856:Person {"name":"Alice"})

    """

    safe_chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_"

    default_sequence_separator = ","
    default_key_value_separator = ":"
    default_element_separator = "\n"

    def __init__(self, file=None, **kwargs):
        Writer.__init__(self, file)
        self.sequence_separator = kwargs.get("sequence_separator", self.default_sequence_separator)
        self.key_value_separator = \
            kwargs.get("key_value_separator", self.default_key_value_separator)
        self.element_separator = kwargs.get("element_separator", self.default_element_separator)

    def write(self, obj):
        """ Write any entity, value or collection.
        """
        if obj is None:
            pass
        elif isinstance(obj, Node):
            self.write_node(id(obj), obj.labels, obj.properties)
        elif isinstance(obj, Rev):
            self.write_rev(obj.type, obj.properties)
        elif isinstance(obj, Rel):
            self.write_rel(obj.type, obj.properties)
        elif isinstance(obj, Path):
            self.write_path(obj)
        elif isinstance(obj, dict):
            self.write_map(obj)
        elif is_collection(obj):
            self.write_list(obj)
        else:
            self.write_value(obj)

    def write_value(self, value):
        """ Write a value.
        """
        self.file.write(ustr(json.dumps(value, ensure_ascii=False)))

    def write_identifier(self, identifier):
        """ Write an identifier.
        """
        if not identifier:
            raise ValueError("Invalid identifier")
        identifier = ustr(identifier)
        safe = all(ch in self.safe_chars for ch in identifier)
        if not safe:
            self.file.write(json.dumps(identifier))
        else:
            self.file.write(identifier)

    def write_list(self, collection):
        """ Write a list.
        """
        self.file.write("[")
        link = ""
        for value in collection:
            self.file.write(link)
            self.write(value)
            link = self.sequence_separator
        self.file.write("]")

    def write_map(self, mapping):
        """ Write a map.
        """
        self.file.write("{")
        link = ""
        for key, value in sorted(mapping.items()):
            self.file.write(link)
            self.write_value(key)
            self.file.write(self.key_value_separator)
            self.write(value)
            link = self.sequence_separator
        self.file.write("}")

    def write_node(self, name, labels=None, properties=None, unique_label=None, unique_key=None):
        """ Write a node.
        """
        self.file.write("(")
        if name:
            self.write_identifier(name)
        for label in sorted(labels or []):
            self.file.write(":")
            self.write_identifier(label)
            if label == unique_label:
                self.file.write("!")
                if unique_key:
                    self.file.write(unique_key)
        if properties is not None:
            if name or labels:
                self.file.write(" ")
            self.write_map(properties)
        self.file.write(")")

    def write_rel(self, type_, properties=None, unique=False):
        """ Write a forward relationship (excluding nodes).
        """
        self.file.write("-[:")
        self.write_identifier(type_)
        if unique:
            self.file.write("!")
        if properties is not None:
            self.file.write(" ")
            self.write_map(properties)
        self.file.write("]->")

    def write_rev(self, type_, properties=None, unique=False):
        """ Write a reverse relationship (excluding nodes).
        """
        self.file.write("<-[:")
        self.write_identifier(type_)
        if unique:
            self.file.write("!")
        if properties is not None:
            self.file.write(" ")
            self.write_map(properties)
        self.file.write("]-")

    def write_path(self, path):
        """ Write a :class:`py2neo.Path`.
        """
        nodes = path.nodes
        self.write_node(id(nodes[0]))
        for i, rel in enumerate(path.rels):
            if isinstance(rel, Rev):
                self.write_rev(rel.type, rel.properties)
            else:
                self.write_rel(rel.type, rel.properties)
            self.write_node(id(nodes[i + 1]))

    def write_subgraph(self, subgraph):
        """ Write a :class:`py2neo.Subgraph`.
        """
        # TODO: uniqueness (in Subgraph)
        link = ""
        for node in subgraph.nodes:
            self.file.write(link)
            self.write_node(id(node), node.labels, node.properties)
            link = self.element_separator
        for relationship in subgraph.relationships:
            self.file.write(link)
            self.write_path(relationship)
            link = self.element_separator



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
An implementation of URIs from RFC 3986 (URI Generic Syntax).

See: http://www.ietf.org/rfc/rfc3986.txt
"""


from __future__ import unicode_literals

import re
try:
    from urllib.parse import quote, unquote
except ImportError:
    from urllib import quote, unquote

from .kvlist import KeyValueList
from .util import bstr, ustr, xstr


__all__ = ["general_delimiters", "subcomponent_delimiters",
           "reserved", "unreserved", "percent_encode", "percent_decode",
           "ParameterString", "Authority", "Path", "Query", "URI"]


# RFC 3986 § 2.2.
general_delimiters = ":/?#[]@"
subcomponent_delimiters = "!$&'()*+,;="
reserved = general_delimiters + subcomponent_delimiters

# RFC 3986 § 2.3.
unreserved = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"
              "abcdefghijklmnopqrstuvwxyz"
              "0123456789-._~")


# RFC 3986 § 2.1.
def percent_encode(data, safe=None):
    """ Percent encode a string of data, optionally keeping certain characters
    unencoded.

    """
    if data is None:
        return None
    if isinstance(data, (tuple, list, set)):
        return "&".join(
            percent_encode(value, safe=safe)
            for value in data
        )
    if isinstance(data, dict):
        return "&".join(
            key + "=" + percent_encode(value, safe=safe)
            for key, value in data.items()
        )
    return quote(bstr(data), safe or b"")


def percent_decode(data):
    """ Percent decode a string of data.

    """
    if data is None:
        return None
    return unquote(xstr(data))


class Part(object):
    """ Internal base class for all URI component parts.
    """

    @classmethod
    def _cast(cls, obj):
        """ Convert the object supplied to an instance of this class, if
        possible.
        """
        if obj is None:
            return cls(None)
        elif isinstance(obj, cls):
            return obj
        else:
            return cls(ustr(obj))

    def __init__(self):
        pass

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, repr(self.string))

    def __str__(self):
        return self.string or ""

    def __eq__(self, other):
        if other is None:
            return self.string is None
        try:
            other_string = other.string
        except AttributeError:
            other_string = self._cast(other).string
        return self.string == other_string

    def __ne__(self, other):
        return not self.__eq__(other)

    def __bool__(self):
        return bool(self.string)

    def __nonzero__(self):
        return bool(self.string)

    def __len__(self):
        return len(ustr(self))

    def __iter__(self):
        return iter(self.string)

    @property
    def string(self):
        raise NotImplementedError()


class ParameterString(Part):

    __string = NotImplemented

    def __init__(self, string, separator):
        super(ParameterString, self).__init__()
        self.__separator = separator
        self.__none = string is None
        self.__parameters = KeyValueList()
        if string:
            bits = string.split(self.__separator)
            for bit in bits:
                if "=" in bit:
                    key, value = map(percent_decode, bit.partition("=")[0::2])
                else:
                    key, value = percent_decode(bit), None
                self.__parameters.append(key, value)

    def __len__(self):
        return self.__parameters.__len__()

    def __bool__(self):
        return bool(self.__parameters)

    def __nonzero__(self):
        return bool(self.__parameters)

    def __contains__(self, item):
        return self.__parameters.__contains__(item)

    def __hash__(self):
        return hash(self.string)

    def __iter__(self):
        return self.__parameters.__iter__()

    def __getitem__(self, index):
        if isinstance(index, slice):
            out = ParameterString("", self.__separator)
            out.__parameters.extend(self.__parameters.__getitem__(index))
            return out
        else:
            return self.__parameters.__getitem__(index)

    def __getslice__(self, start, stop):
        out = ParameterString("", self.__separator)
        out.__parameters.extend(self.__parameters.__getslice__(start, stop))
        return out

    def get(self, name, index=0):
        if not self.__parameters.has_key(name):
            raise KeyError(name)
        for i, value in enumerate(self.__parameters.get(name)):
            if i == index:
                return value
        raise IndexError("Parameter {0} does not have {1} "
                         "values".format(name, index))

    def get_all(self, name):
        if not self.__parameters.has_key(name):
            raise KeyError(name)
        return list(self.__parameters.get(name))

    @property
    def string(self):
        if self.__string is NotImplemented:
            if self.__none:
                self.__string = None
            else:
                bits = []
                for key, value in self.__parameters:
                    if value is None:
                        bits.append(percent_encode(key))
                    else:
                        bits.append(percent_encode(key) + "=" + percent_encode(value))
                self.__string = self.__separator.join(bits)
        return self.__string


class Authority(Part):
    """ A host name plus optional port and user information detail.

    **Syntax**
        ``authority := [ user_info "@" ] host [ ":" port ]``

    .. seealso::
        `RFC 3986 § 3.2`_

    .. _`RFC 3986 § 3.2`: http://tools.ietf.org/html/rfc3986#section-3.2
    """

    @classmethod
    def _parse_host_port(cls, string):
        host, colon, port = string.partition(":")
        if colon:
            return host, int(port)
        else:
            return host, None

    __instances = {}

    def __new__(cls, string):
        try:
            inst = cls.__instances[string]
        except KeyError:
            inst = super(Authority, cls).__new__(cls)
            if string is not None:
                user_info, at, host_port = string.rpartition("@")
                if at:
                    inst.__user_info = percent_decode(user_info)
                inst.__host, inst.__port = cls._parse_host_port(host_port)
            cls.__instances[string] = inst
        return inst

    __user_info = None
    __host = None
    __port = None

    __string = NotImplemented

    def __init__(self, string):
        super(Authority, self).__init__()

    def __hash__(self):
        return hash(self.string)

    @property
    def host(self):
        """ The host part of this authority component, an empty string if host
        is empty or :py:const:`None` if undefined.

        ::

            >>> Authority(None).host
            None
            >>> Authority("").host
            ''
            >>> Authority("example.com").host
            'example.com'
            >>> Authority("example.com:8080").host
            'example.com'
            >>> Authority("bob@example.com").host
            'example.com'
            >>> Authority("bob@example.com:8080").host
            'example.com'

        :return:
        """
        return self.__host

    @property
    def host_port(self):
        """ The host and port parts of this authority component or
        :py:const:`None` if undefined.

        ::

            >>> Authority(None).host_port
            None
            >>> Authority("").host_port
            ''
            >>> Authority("example.com").host_port
            'example.com'
            >>> Authority("example.com:8080").host_port
            'example.com:8080'
            >>> Authority("bob@example.com").host_port
            'example.com'
            >>> Authority("bob@example.com:8080").host_port
            'example.com:8080'

        :return:
        """
        u = [self.__host]
        if self.__port is not None:
            u += [":", ustr(self.__port)]
        return "".join(u)

    @property
    def port(self):
        """ The port part of this authority component or :py:const:`None` if
        undefined.

        ::

            >>> Authority(None).port
            None
            >>> Authority("").port
            None
            >>> Authority("example.com").port
            None
            >>> Authority("example.com:8080").port
            8080
            >>> Authority("bob@example.com").port
            None
            >>> Authority("bob@example.com:8080").port
            8080

        :return:
        """
        return self.__port

    @property
    def string(self):
        """ The full string value of this authority component or
        :`py:const:`None` if undefined.

        ::

            >>> Authority(None).string
            None
            >>> Authority("").string
            ''
            >>> Authority("example.com").string
            'example.com'
            >>> Authority("example.com:8080").string
            'example.com:8080'
            >>> Authority("bob@example.com").string
            'bob@example.com'
            >>> Authority("bob@example.com:8080").string
            'bob@example.com:8080'

        :return:
        """
        if self.__string is NotImplemented:
            if self.__host is None:
                self.__string = None
            else:
                u = []
                if self.__user_info is not None:
                    u += [percent_encode(self.__user_info), "@"]
                u += [self.__host]
                if self.__port is not None:
                    u += [":", ustr(self.__port)]
                self.__string = "".join(u)
        return self.__string
        
    @property
    def user_info(self):
        """ The user information part of this authority component or
        :py:const:`None` if undefined.

        ::

            >>> Authority(None).user_info
            None
            >>> Authority("").user_info
            None
            >>> Authority("example.com").user_info
            None
            >>> Authority("example.com:8080").user_info
            None
            >>> Authority("bob@example.com").user_info
            'bob'
            >>> Authority("bob@example.com:8080").user_info
            'bob'

        :return:
        """
        return self.__user_info


class Path(Part):

    __segments = None
    __string = NotImplemented

    def __init__(self, string):
        super(Path, self).__init__()
        if string is not None:
            self.__segments = tuple(map(percent_decode, string.split("/")))

    def __hash__(self):
        return hash(self.string)

    @property
    def string(self):
        if self.__string is NotImplemented:
            if self.__segments is None:
                self.__string = None
            else:
                self.__string = "/".join(map(percent_encode, self.__segments))
        return self.__string

    @property
    def segments(self):
        if self.__segments is None:
            return ()
        else:
            return self.__segments

    def __iter__(self):
        return iter(self.__segments)

    def remove_dot_segments(self):
        """ Implementation of RFC3986, section 5.2.4
        """
        inp = self.string
        out = ""
        while inp:
            if inp.startswith("../"):
                inp = inp[3:]
            elif inp.startswith("./"):
                inp = inp[2:]
            elif inp.startswith("/./"):
                inp = inp[2:]
            elif inp == "/.":
                inp = "/"
            elif inp.startswith("/../"):
                inp = inp[3:]
                out = out.rpartition("/")[0]
            elif inp == "/..":
                inp = "/"
                out = out.rpartition("/")[0]
            elif inp in (".", ".."):
                inp = ""
            else:
                if inp.startswith("/"):
                    inp = inp[1:]
                    out += "/"
                seg, slash, inp = inp.partition("/")
                out += seg
                inp = slash + inp
        return Path(out)

    def with_trailing_slash(self):
        if self.__segments is None:
            return self
        s = self.string
        if s.endswith("/"):
            return self
        else:
            return Path(s + "/")

    def without_trailing_slash(self):
        if self.__segments is None:
            return self
        s = self.string
        if s.endswith("/"):
            return Path(s[:-1])
        else:
            return self


class Query(ParameterString):

    SEPARATOR = "&"

    def __init__(self, string):
        super(Query, self).__init__(string, self.SEPARATOR)

    def __hash__(self):
        return hash(self.string)


class URI(Part):
    """ Uniform Resource Identifier.

    .. seealso::
        `RFC 3986`_

    .. _`RFC 3986`: http://tools.ietf.org/html/rfc3986
    """

    @classmethod
    def build(cls, **parts):
        """ Build a URI object from named parts. The part names available are:

        - string
        - hierarchical_part
        - absolute_path_reference
        - authority
        - host_port
        - scheme
        - user_info
        - host
        - port
        - path
        - query
        - fragment

        """
        uri = URI(parts.get("string"))
        uri.__set_hierarchical_part(parts.get("hierarchical_part"))
        uri.__set_absolute_path_reference(parts.get("absolute_path_reference"))
        uri.__set_authority(parts.get("authority"))
        uri.__set_host_port(parts.get("host_port"))
        uri.__set_scheme(parts.get("scheme"))
        uri.__set_user_info(parts.get("user_info"))
        uri.__set_host(parts.get("host"))
        uri.__set_port(parts.get("port"))
        uri.__set_path(parts.get("path"))
        uri.__set_query(parts.get("query"))
        uri.__set_fragment(parts.get("fragment"))
        if parts and uri.__path is None:
            uri.__path = Path("")
        return uri

    @classmethod
    def _partition_fragment(cls, value):
        value, hash, fragment = value.partition("#")
        if fragment:
            return value, percent_decode(fragment)
        else:
            return value, None

    @classmethod
    def _partition_query(cls, value):
        value, question_mark, query = value.partition("?")
        if query:
            return value, Query(query)
        else:
            return value, None

    @classmethod
    def _parse_hierarchical_part(cls, value):
        if value.startswith("//"):
            authority, slash, path = value[2:].partition("/")
            if slash:
                return Authority(authority), Path(slash + path)
            else:
                return Authority(authority), Path("")
        else:
            return None, Path(value)

    def __new__(cls, value):
        if isinstance(value, cls):
            return value
        inst = super(cls, URI).__new__(cls)
        if value is None:
            return inst
        try:
            if value.__uri__ is None:
                return inst
        except AttributeError:
            pass
        try:
            value = ustr(value.__uri__)
        except AttributeError:
            value = ustr(value)
        # scheme
        if ":" in value:
            inst.__scheme, _, value = value.partition(":")
            inst.__scheme = percent_decode(inst.__scheme)
        else:
            inst.__scheme = None
        # fragment
        value, inst.__fragment = cls._partition_fragment(value)
        # query
        value, inst.__query = cls._partition_query(value)
        # hierarchical part
        inst.__authority, inst.__path = cls._parse_hierarchical_part(value)
        return inst

    __scheme = None
    __authority = None
    __path = None
    __query = None
    __fragment = None

    __string = NotImplemented

    def __init__(self, value):
        Part.__init__(self)

    def __hash__(self):
        return hash(self.string)

    def __add__(self, other):
        return URI("{0}{1}".format(self, other))

    @property
    def __uri__(self):
        return self.string

    def __set_hierarchical_part(self, string):
        if string is not None:
            self.__authority, self.__path = self._parse_hierarchical_part(string)

    def __set_absolute_path_reference(self, string):
        if string is not None:
            string, self.__fragment = self._partition_fragment(string)
            string, self.__query = self._partition_query(string)
            self.__path = Path(string)

    def __set_authority(self, string):
        if string is not None:
            self.__authority = Authority(string)

    def __set_host_port(self, string):
        if string is not None:
            if self.__authority is None:
                self.__authority = Authority(string)
            else:
                host, port = Authority._parse_host_port(string)
                self.__authority._Authority__host = host
                self.__authority._Authority__port = port

    def __set_scheme(self, string):
        if string is not None:
            self.__scheme = string

    def __set_user_info(self, string):
        if string is not None:
            if self.__authority is None:
                self.__authority = Authority("")
            self.__authority._Authority__user_info = string

    def __set_host(self, string):
        if string is not None:
            if self.__authority is None:
                self.__authority = Authority(string)
            else:
                self.__authority._Authority__host = string

    def __set_port(self, number):
        if number is not None:
            if self.__authority is None:
                self.__authority = Authority("")
            self.__authority._Authority__port = number

    def __set_path(self, string):
        if string is not None:
            self.__path = Path(string)

    def __set_query(self, string):
        if string is not None:
            self.__query = Query(string)

    def __set_fragment(self, string):
        if string is not None:
            self.__fragment = string

    @property
    def string(self):
        """ The full percent-encoded string value of this URI or
        :py:const:`None` if undefined.

        ::

            >>> URI(None).string
            None
            >>> URI("").string
            ''
            >>> URI("http://example.com").string
            'example.com'
            >>> URI("foo/bar").string
            'foo/bar'
            >>> URI("http://bob@example.com:8080/data/report.html?date=2000-12-25#summary").string
            'http://bob@example.com:8080/data/report.html?date=2000-12-25#summary'

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
            \___________________________________________________________________/
                                             |
                                           string

        :rtype: percent-encoded string or :py:const:`None`

        .. note::
            Unlike ``string``, the ``__str__`` method will always return a
            string, even when the URI is undefined; in this case, an empty
            string is returned instead of :py:const:`None`.
        """
        if self.__string is NotImplemented:
            if self.__path is None:
                self.__string = None
            else:
                u = []
                if self.__scheme is not None:
                    u += [percent_encode(self.__scheme), ":"]
                if self.__authority is not None:
                    u += ["//", ustr(self.__authority)]
                u += [ustr(self.__path)]
                if self.__query is not None:
                    u += ["?", ustr(self.__query)]
                if self.__fragment is not None:
                    u += ["#", percent_encode(self.__fragment)]
                self.__string = "".join(u)
        return self.__string

    @property
    def scheme(self):
        """ The scheme part of this URI or :py:const:`None` if undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
            \___/
              |
            scheme

        :rtype: unencoded string or :py:const:`None`
        """
        return self.__scheme

    @property
    def authority(self):
        """ The authority part of this URI or :py:const:`None` if undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                    \__________________/
                             |
                         authority

        :rtype: :py:class:`Authority <httpstream.uri.Authority>` instance or
            :py:const:`None`
        """
        return self.__authority

    @property
    def user_info(self):
        """ The user information part of this URI or :py:const:`None` if
        undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                    \_/
                     |
                 user_info

        :return: string value of user information part or :py:const:`None`
        :rtype: unencoded string or :py:const:`None`
        """
        if self.__authority is None:
            return None
        else:
            return self.__authority.user_info

    @property
    def host(self):
        """ The *host* part of this URI or :py:const:`None` if undefined.

        ::

            >>> URI(None).host
            None
            >>> URI("").host
            None
            >>> URI("http://example.com").host
            'example.com'
            >>> URI("http://example.com:8080/data").host
            'example.com'

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                        \_________/
                             |
                            host

        :return:
        :rtype: unencoded string or :py:const:`None`
        """
        if self.__authority is None:
            return None
        else:
            return self.__authority.host
        
    @property
    def port(self):
        """ The *port* part of this URI or :py:const:`None` if undefined.

        ::

            >>> URI(None).port
            None
            >>> URI("").port
            None
            >>> URI("http://example.com").port
            None
            >>> URI("http://example.com:8080/data").port
            8080

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                                    \__/
                                     |
                                    port

        :return:
        :rtype: integer or :py:const:`None`
        """
        if self.__authority is None:
            return None
        else:
            return self.__authority.port

    @property
    def host_port(self):
        """ The *host* and *port* parts of this URI separated by a colon or
        :py:const:`None` if both are undefined.

        ::

            >>> URI(None).host_port
            None
            >>> URI("").host_port
            None
            >>> URI("http://example.com").host_port
            'example.com'
            >>> URI("http://example.com:8080/data").host_port
            'example.com:8080'
            >>> URI("http://bob@example.com:8080/data").host_port
            'example.com:8080'

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                        \______________/
                               |
                           host_port

        :return:
        :rtype: percent-encoded string or :py:const:`None`
        """
        if self.__authority is None:
            return None
        else:
            return self.__authority.host_port

    @property
    def path(self):
        """ The *path* part of this URI or :py:const:`None` if undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                                        \_______________/
                                                |
                                               path

        :return:
        :rtype: :py:class:`Path <httpstream.uri.Path>` instance or
            :py:const:`None`
        """
        return self.__path

    @property
    def query(self):
        """ The *query* part of this URI or :py:const:`None` if undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                                                          \_____________/
                                                                 |
                                                               query

        :rtype: :py:class:`Query <httpstream.uri.Query>` instance or
            :py:const:`None`
        """
        return self.__query

    @property
    def fragment(self):
        """ The *fragment* part of this URI or :py:const:`None` if undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                                                                          \_____/
                                                                             |
                                                                          fragment

        :return:
        :rtype: unencoded string or :py:const:`None`
        """
        return self.__fragment

    @property
    def hierarchical_part(self):
        """ The authority and path parts of this URI or :py:const:`None` if
        undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                    \___________________________________/
                                      |
                              hierarchical_part

        :return: combined string values of authority and path parts or
            :py:const:`None`
        :rtype: percent-encoded string or :py:const:`None`
        """
        if self.__path is None:
            return None
        u = []
        if self.__authority is not None:
            u += ["//", ustr(self.__authority)]
        u += [ustr(self.__path)]
        return "".join(u)

    @property
    def absolute_path_reference(self):
        """ The path, query and fragment parts of this URI or :py:const:`None`
        if undefined.

        **Component Definition:**
        ::

            https://bob@example.com:8080/data/report.html?date=2000-12-25#summary
                                        \_______________________________________/
                                                            |
                                                  absolute_path_reference

        :return: combined string values of path, query and fragment parts or
            :py:const:`None`
        :rtype: percent-encoded string or :py:const:`None`
        """
        if self.__path is None:
            return None
        u = [ustr(self.__path)]
        if self.__query is not None:
            u += ["?", ustr(self.__query)]
        if self.__fragment is not None:
            u += ["#", percent_encode(self.__fragment)]
        return "".join(u)

    def _merge_path(self, relative_path_reference):
        if self.__authority is not None and not self.__path:
            return Path("/" + ustr(relative_path_reference))
        elif "/" in self.__path.string:
            segments = self.__path.segments[:-1] + ("",)
            return Path("/".join(segments) + ustr(relative_path_reference))
        else:
            return relative_path_reference

    def resolve(self, reference, strict=True):
        """ Transform a reference relative to this URI to produce a full target
        URI.

        .. seealso::
            `RFC 3986 § 5.2.2`_

        .. _`RFC 3986 § 5.2.2`: http://tools.ietf.org/html/rfc3986#section-5.2.2
        """
        if reference is None:
            return None
        reference = self._cast(reference)
        target = URI(None)
        if not strict and reference.__scheme == self.__scheme:
            reference_scheme = None
        else:
            reference_scheme = reference.__scheme
        if reference_scheme is not None:
            target.__scheme = reference_scheme
            target.__authority = reference.__authority
            target.__path = reference.__path.remove_dot_segments()
            target.__query = reference.__query
        else:
            if reference.__authority is not None:
                target.__authority = reference.__authority
                target.__path = reference.__path.remove_dot_segments()
                target.__query = reference.__query
            else:
                if not reference.path:
                    target.__path = self.__path
                    if reference.__query is not None:
                        target.__query = reference.__query
                    else:
                        target.__query = self.__query
                else:
                    if ustr(reference.__path).startswith("/"):
                        target.__path = reference.__path.remove_dot_segments()
                    else:
                        target.__path = self._merge_path(reference.__path)
                        target.__path = target.__path.remove_dot_segments()
                    target.__query = reference.__query
                target.__authority = self.__authority
            target.__scheme = self.__scheme
        target.__fragment = reference.__fragment
        return target



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo import Service, ResourceTemplate, GraphError
from py2neo.packages.httpstream import Response
from py2neo.packages.httpstream.numbers import NOT_FOUND


__all__ = ["SchemaResource"]


class SchemaResource(Service):
    """ The schema resource attached to a `Graph` instance.
    """

    __instances = {}

    def __new__(cls, uri):
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(SchemaResource, cls).__new__(cls)
            inst.bind(uri)
            if not inst.graph.supports_schema_indexes:
                raise NotImplementedError("Schema index support requires version 2.0 or above")
            inst._index_template = ResourceTemplate(uri + "/index/{label}")
            inst._index_key_template = ResourceTemplate(uri + "/index/{label}/{property_key}")
            inst._uniqueness_constraint_template = \
                ResourceTemplate(uri + "/constraint/{label}/uniqueness")
            inst._uniqueness_constraint_key_template = \
                ResourceTemplate(uri + "/constraint/{label}/uniqueness/{property_key}")
            cls.__instances[uri] = inst
        return inst

    def create_index(self, label, property_key):
        """ Create a schema index for a label and property
        key combination.
        """
        self._index_template.expand(label=label).post({"property_keys": [property_key]})

    def create_uniqueness_constraint(self, label, property_key):
        """ Create a uniqueness constraint for a label.
        """
        self._uniqueness_constraint_template.expand(label=label).post(
            {"property_keys": [property_key]})

    def drop_index(self, label, property_key):
        """ Remove label index for a given property key.
        """
        try:
            self._index_key_template.expand(label=label, property_key=property_key).delete()
        except GraphError as error:
            cause = error.__cause__
            if isinstance(cause, Response):
                if cause.status_code == NOT_FOUND:
                    raise GraphError("No such schema index (label=%r, key=%r)" % (
                        label, property_key))
            raise

    def drop_uniqueness_constraint(self, label, property_key):
        """ Remove the uniqueness constraint for a given property key.
        """
        try:
            self._uniqueness_constraint_key_template.expand(
                label=label, property_key=property_key).delete()
        except GraphError as error:
            cause = error.__cause__
            if isinstance(cause, Response):
                if cause.status_code == NOT_FOUND:
                    raise GraphError("No such unique constraint (label=%r, key=%r)" % (
                        label, property_key))
            raise

    def get_indexes(self, label):
        """ Fetch a list of indexed property keys for a label.
        """
        return [
            indexed["property_keys"][0]
            for indexed in self._index_template.expand(label=label).get().content
        ]

    def get_uniqueness_constraints(self, label):
        """ Fetch a list of unique constraints for a label.
        """
        return [
            unique["property_keys"][0]
            for unique in self._uniqueness_constraint_template.expand(label=label).get().content
        ]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function

import base64
import os
import sys

from py2neo import GraphError, Service, ServiceRoot
from py2neo.env import NEO4J_URI
from py2neo.packages.httpstream.numbers import UNPROCESSABLE_ENTITY
from py2neo.util import ustr


HELP = """\
Usage: {script} «user_name» «password» [«new_password»]

Authenticate against a Neo4j database server, optionally changing
the password. If no new password is supplied, a simple auth check
is carried out and a response is output describing whether or not
a password change is required. If the new password is supplied, a
password change is attempted.

Report bugs to nigel@py2neo.org
"""


def auth_header_value(user_name, password, realm=None):
    """ Construct a value for the Authorization header based on the
    credentials supplied.

    :param user_name: the user name
    :param password: the password
    :param realm: the realm (optional)
    :return: string to be included in Authorization header
    """
    credentials = (user_name + ":" + password).encode("UTF-8")
    if realm:
        value = 'Basic realm="' + realm + '" '
    else:
        value = 'Basic '
    value += base64.b64encode(credentials).decode("ASCII")
    return value


class UserManager(Service):
    """ User management service.
    """

    @classmethod
    def for_user(cls, service_root, user_name, password):
        """ Fetch a UserManager instance for a given service root and user name.

        :param service_root: A valid :class:`py2neo.ServiceRoot` instance.
        :rtype: :class:`.UserManager`
        """
        uri = service_root.uri.resolve("/user/%s" % user_name)
        inst = cls(uri)
        inst.resource._headers["Authorization"] = auth_header_value(user_name, password, "Neo4j")
        return inst

    __instances = {}

    #: Instance metadata, updated on refresh.
    metadata = None

    def __new__(cls, uri):
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(UserManager, cls).__new__(cls)
            inst.bind(uri)
            cls.__instances[uri] = inst
        return inst

    def refresh(self):
        """ Refresh the stored metadata.
        """
        rs = self.resource.get()
        self.metadata = rs.content

    @property
    def user_name(self):
        self.refresh()
        return self.metadata["username"]

    @property
    def password_manager(self):
        self.refresh()
        password_manager = PasswordManager(self.metadata["password_change"])
        password_manager.resource._headers["Authorization"] = \
            self.resource._headers["Authorization"]
        return password_manager

    @property
    def password_change_required(self):
        self.refresh()
        return self.metadata["password_change_required"]


class PasswordManager(Service):
    """ Password management service.
    """

    __instances = {}

    def __new__(cls, uri):
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(PasswordManager, cls).__new__(cls)
            inst.bind(uri)
            cls.__instances[uri] = inst
        return inst

    def change(self, new_password):
        """ Change the authentication password.

        :arg password: The current password.
        :arg new_password: A new password (must not match the current password).
        :return: A valid auth token.
        :raise ValueError: If the new password is invalid.
        """
        try:
            response = self.resource.post({"password": new_password})
        except GraphError as error:
            if error.response.status_code == UNPROCESSABLE_ENTITY:
                raise ValueError("Cannot change password")
            else:
                raise
        else:
            return response.status_code == 200


def _help(script):
    print(HELP.format(script=os.path.basename(script)))


def main():
    script, args = sys.argv[0], sys.argv[1:]
    if len(args) < 2 or len(args) > 3:
        _help(script)
        return
    try:
        service_root = ServiceRoot(NEO4J_URI)
        user_name = args[0]
        password = args[1]
        user_manager = UserManager.for_user(service_root, user_name, password)
        if len(args) == 2:
            # Check password
            if user_manager.password_change_required:
                print("Password change required")
            else:
                print("Password change not required")
        else:
            # Change password
            password_manager = user_manager.password_manager
            new_password = args[2]
            if password_manager.change(new_password):
                print("Password change succeeded")
            else:
                print("Password change failed")
                sys.exit(2)
    except Exception as error:
        sys.stderr.write("%s: %s\n" % (error.__class__.__name__, ustr(error)))
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.batch.core import Batch


__all__ = ["ReadBatch"]


class ReadBatch(Batch):
    """ Generic batch execution facility for data read requests,
    """

    def __init__(self, graph):
        Batch.__init__(self, graph)

    def stream(self):
        for result in self.graph.batch.stream(self):
            yield result.content

    def submit(self):
        return [result.content for result in self.graph.batch.submit(self)]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys

try:
    from setuptools import setup, find_packages
    from setuptools.extension import Extension
except ImportError:
    from distutils.core import setup, find_packages
    from distutils.extension import Extension

from py2neo import __author__, __email__, __license__, __package__, __version__


python_2 = sys.version_info < (3,)

packages = find_packages(exclude=("book", "examples", "examples.*", "test", "test.*"))
package_metadata = {
    "name": "py2neo-archive",
    "version": __version__,
    "description": "Python client library and toolkit for Neo4j",
    "long_description": "Py2neo is a client library and comprehensive toolkit for working with "
                        "Neo4j from within Python applications and from the command line. The "
                        "core library has no external dependencies and has been carefully "
                        "designed to be easy and intuitive to use.",
    "author": __author__,
    "author_email": __email__,
    "url": "http://py2neo.org/",
    "entry_points": {
        "console_scripts": [
            "cypher = py2neo.cypher.__main__:main",
            "neoauth = py2neo.password:main",
            "neoget = py2neo.server:main",
            "geoff = py2neo.ext.geoff.__main__:main",
            "neobox = py2neo.ext.neobox.__main__:main",
        ],
    },
    "packages": packages,
    "license": __license__,
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Database",
        "Topic :: Software Development",
    ],
    "zip_safe": False,
}

extensions = []
sdist = "sdist" in sys.argv
if sdist or not python_2:
    extensions.append(Extension("py2neo.packages.jsonstream.cjsonstream",
                                ["py2neo/packages/jsonstream/cjsonstream.c"]))
if sdist or python_2:
    extensions.append(Extension("py2neo.packages.jsonstream.cjsonstream_2x",
                                ["py2neo/packages/jsonstream/cjsonstream_2x.c"]))

try:
    setup(ext_modules=extensions, **package_metadata)
except:
    setup(**package_metadata)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.core import Graph, Node, Path, Relationship
from py2neo.cypher.util import StartOrMatch
from py2neo.util import ustr, xstr


__all__ = ["DeleteStatement"]


def _(*args):
    return "".join("_" + ustr(arg) for arg in args)


class DeleteStatement(object):
    """ Builder for a Cypher DELETE statement.
    """

    #: The graph against which this statement is to be executed.
    graph = None

    #: The parameters to inject into this statement.
    parameters = None

    def __init__(self, graph):
        self.graph = graph
        self.supports_node_labels = self.graph.supports_node_labels
        self.entities = []
        self.start_or_match_clause = StartOrMatch(self.graph)
        self.delete_rels_clause = []
        self.delete_nodes_clause = []
        self.parameters = {}

    def __repr__(self):
        return self.string

    def __str__(self):
        return xstr(self.__unicode__())

    def __unicode__(self):
        return self.string

    def __contains__(self, entity):
        return any(e is entity for e in self.entities)

    @property
    def string(self):
        """ The full Cypher statement as a string.
        """
        clauses = []
        if self.start_or_match_clause:
            clauses.append(self.start_or_match_clause.string.rstrip())
        if self.delete_rels_clause:
            clauses.append("DELETE " + ",".join(self.delete_rels_clause))
        if self.delete_nodes_clause:
            clauses.append("DELETE " + ",".join(self.delete_nodes_clause))
        return "\n".join(clauses)

    def post(self):
        return self.graph.cypher.post(self.string, self.parameters)

    def execute(self):
        """ Execute this statement.
        """
        if self.string:
            self.post().close()

    def delete(self, entity):
        """ Append an entity to the DELETE clause of this statement.

        :arg entity: The entity to delete.

        """
        entity = Graph.cast(entity)
        index = len(self.entities)
        name = _(index)
        if isinstance(entity, Node):
            self._delete_node(entity, name)
        elif isinstance(entity, Relationship):
            self._delete_relationship(entity, name)
        elif isinstance(entity, Path):
            self._delete_path(entity, name)
        self.entities.append(entity)

    def _delete_node(self, node, name):
        if node.bound:
            self.start_or_match_clause.node(name, "{%s}" % name)
            self.delete_nodes_clause.append(name)
            self.parameters[name] = node._id

    def _delete_relationship(self, relationship, name):
        if relationship.bound:
            self.start_or_match_clause.relationship(name, "{%s}" % name)
            self.delete_rels_clause.append(name)
            self.parameters[name] = relationship._id

    def _delete_path(self, path, name):
        for i, rel in enumerate(path.relationships):
            self._delete_relationship(rel, name + "r" + ustr(i))
        for i, node in enumerate(path.nodes):
            self._delete_node(node, name + "n" + ustr(i))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2012-2014 Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

from datetime import date, time, datetime
from decimal import Decimal
import json


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, (datetime, date, time)):
            return obj.isoformat()
        if isinstance(obj, Decimal):
            return str(obj)
        if isinstance(obj, (set, frozenset)):
            return list(obj)
        if isinstance(obj, complex):
            return [obj.real, obj.imag]
        return json.JSONEncoder.default(self, obj)
